#################################################################################################
# Copyright (c) 2022 - 2026 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################

import functools
from typing import List, Tuple

import torch
from torch import Tensor
from torch.amp import custom_bwd, custom_fwd
from torch.autograd import Function

amp_fwd = functools.partial(custom_fwd, device_type="cuda")
amp_bwd = functools.partial(custom_bwd, device_type="cuda")

from natten._environment import _IS_TORCH_COMPILE_SUPPORTED


def _maybe_torch_compile(*args, **kwargs):
    def decorator(f):
        if _IS_TORCH_COMPILE_SUPPORTED:
            return torch.compile(f, *args, **kwargs)
        return f

    return decorator


# TODO: if use cases for this grow, we might want to do a custom kernel
def _merge_attentions_fn(
    outputs: List[Tensor], lse_tensors: List[Tensor]
) -> Tuple[Tensor, Tensor]:

    assert len(outputs) >= 2, "Expected at least two tensors."
    num_splits = len(outputs)
    assert (
        len(lse_tensors) == num_splits
    ), "Expected number of outputs and LSE tensors to match."

    assert all(
        output.dim() == 4 and output.is_contiguous() for output in outputs
    ), "Output tensors must be rank-4 tensors with (batch, seq, heads, dim) contiguous layout."

    batch, seqlen, heads, dim = outputs[0].shape

    assert all(
        [x for x in output.shape] == [batch, seqlen, heads, dim] for output in outputs
    ), "Output tensors must match in shape."

    assert all(
        lse.dim() == 3
        and lse.is_contiguous()
        and [x for x in lse.shape] == [batch, seqlen, heads]
        for lse in lse_tensors
    ), "LSE tensors must be rank-3 tensors with (batch, seq, heads) contiguous layout, and match in shape."

    accum_type = torch.float32
    output_type = outputs[0].dtype

    lse_tensors = [lse.to(accum_type).unsqueeze(-1) for lse in lse_tensors]

    outputs = [output.to(accum_type) for output in outputs]

    # New approach based on https://github.com/zhuzilin/ring-flash-attention/pull/34
    output = outputs[0] - torch.nn.functional.sigmoid(
        lse_tensors[1] - lse_tensors[0]
    ) * (outputs[0] - outputs[1])
    logsumexp = lse_tensors[0] - torch.nn.functional.logsigmoid(
        lse_tensors[0] - lse_tensors[1]
    )
    for i in range(2, num_splits):
        output = output - torch.nn.functional.sigmoid(lse_tensors[i] - logsumexp) * (
            output - outputs[i]
        )
        logsumexp = logsumexp - torch.nn.functional.logsigmoid(
            logsumexp - lse_tensors[i]
        )

    output = output.to(output_type)
    logsumexp = logsumexp.squeeze(-1)

    assert logsumexp.dim() == 3
    assert logsumexp.shape[0] == batch
    assert logsumexp.shape[1] == seqlen
    assert logsumexp.shape[2] == heads

    return output, logsumexp


@_maybe_torch_compile(fullgraph=True)
def _merge_attentions_compile(
    outputs: List[Tensor], lse_tensors: List[Tensor]
) -> Tuple[Tensor, Tensor]:
    return _merge_attentions_fn(outputs, lse_tensors)


def _merge_attentions_op(
    outputs: List[Tensor], lse_tensors: List[Tensor], torch_compile: bool = True
) -> Tuple[Tensor, Tensor]:

    if not torch_compile:
        return _merge_attentions_fn(
            [output.contiguous() for output in outputs],
            [lse.contiguous() for lse in lse_tensors],
        )

    return _merge_attentions_compile(
        [output.contiguous() for output in outputs],
        [lse.contiguous() for lse in lse_tensors],
    )


class MergeAttentionsAutogradFn(Function):
    @staticmethod
    @amp_fwd
    def forward(
        ctx,
        output_0: Tensor,
        output_1: Tensor,
        lse_0: Tensor,
        lse_1: Tensor,
        torch_compile: bool,
    ) -> Tuple[Tensor, Tensor]:

        merged_output, merged_lse = _merge_attentions_op(
            [output_0.contiguous(), output_1.contiguous()],
            [lse_0.contiguous(), lse_1.contiguous()],
            torch_compile=torch_compile,
        )

        ctx.save_for_backward(
            output_0, output_1, lse_0, lse_1, merged_output, merged_lse
        )

        return merged_output, merged_lse

    @staticmethod
    @amp_bwd
    def backward(
        ctx, grad_out: Tensor, grad_lse: Tensor
    ) -> Tuple[Tensor, Tensor, Tensor, Tensor, None]:

        output_0, output_1, lse_0, lse_1, merged_output, merged_lse = ctx.saved_tensors

        # Outputs and LSEs from the originating attention ops must be replaced with
        # the merged ones inplace so that we get correct behavior, and not break torch.compile
        # graphs in the process.
        output_0.data.copy_(merged_output.data.reshape(output_0.shape))
        output_1.data.copy_(merged_output.data.reshape(output_1.shape))

        lse_0.data.copy_(merged_lse.data.reshape(lse_0.shape))
        lse_1.data.copy_(merged_lse.data.reshape(lse_1.shape))

        return grad_out, grad_out, grad_lse, grad_lse, None


def merge_attentions(
    outputs: List[Tensor],
    lse_tensors: List[Tensor],
    torch_compile: bool = True,
    use_autograd_fix: bool = True,
) -> Tuple[Tensor, Tensor]:
    """Takes multiple attention *outputs* originating from the same query tensor, and their
    corresponding logsumexps, and merges them as if their context (key/value pair) had been
    concatenated.

    This operation is used to implement cross-neighborhood attention, and can also be used for
    distributed setups, such as context-parallelism.

    This operation also attempts to use `torch.compile` to fuse the elementwise operations. This
    can be disabled by passing `torch_compile=False`.

    Parameters:
        outputs (List[Tensor]): List of 4-D attention output tensors, with the heads last layout
            (`[batch, seqlen, heads, head_dim]`)

        lse_tensors (List[Tensor]): List of 3-D logsumexp tensors, with the heads last layout
            (`[batch, seqlen, heads]`)

        torch_compile (bool): Attempt to use `torch.compile` to fuse the underlying elementwise
            operations. Default: True.

        use_autograd_fix (bool): fix backpropagation by using a custom autograd function. Only
            compatible with fused attention operations (Flash/FMHA/FNA), and must be disabled when
            using unfused Attention, which includes Flex without torch.compile.
            This operation only supports backpropagation with pairs (when `len(outputs) == 2`).
            Default: True.

    Returns:
        output (Tensor): merged attention output.

        logsumexp (Tensor): updated logsumexp.
    """

    if len(outputs) < 2:
        raise ValueError("`merge_attentions` expects at least two tensors.")

    if len(outputs) != len(lse_tensors):
        raise ValueError(
            "`merge_attentions` expected number of outputs and LSE tensors to match, "
            f"got {len(outputs)=} != {len(lse_tensors)}."
        )

    requires_grad = outputs[0].requires_grad
    shape = outputs[0].shape

    for i, (output, lse) in enumerate(zip(outputs, lse_tensors)):
        if output.dim() != 4 or not output.is_contiguous():
            raise ValueError(
                "Output tensors must be rank-4 tensors with (batch, seq, heads, dim), "
                f"but got output {i} with rank={output.dim()}."
            )

        if output.shape != shape:
            raise ValueError(
                f"Output tensors must must match in shape, but got output {i} "
                f"with shape={output.shape}."
            )

        if lse.dim() != 3:
            raise ValueError(
                "LSE tensors must be rank-3 tensors with (batch, seq, heads)"
                f"but got LSE {i} with rank={lse.dim()}."
            )

        if lse.shape != shape[:3]:
            raise ValueError(
                f"LSE tensors must must match outputs in shape except last dim "
                f"({shape=}), but got LSE {i} with shape={lse.shape}."
            )

        if output.requires_grad and not requires_grad:
            raise ValueError(
                "Either all attentions must require grad, or none of them."
            )

    # This path is the correct way to do backward pass, but since we can't have lists as inputs to
    # autograd functions, we're forced to specialize it for 2-way for now.
    if use_autograd_fix:
        if requires_grad and len(outputs) == 2:
            assert len(outputs) == len(lse_tensors)

            merged_output, merged_lse = MergeAttentionsAutogradFn.apply(
                outputs[0], outputs[1], lse_tensors[0], lse_tensors[1], torch_compile
            )

            return merged_output, merged_lse

        if requires_grad:
            raise NotImplementedError(
                "'merge_attentions' only supports backwards pass with two inputs, "
                f"got {len(outputs)=}."
            )

    return _merge_attentions_op(
        [output.contiguous() for output in outputs],
        [lse.contiguous() for lse in lse_tensors],
        torch_compile=torch_compile,
    )


__all__ = ["merge_attentions"]



#################################################################################################
# Copyright (c) 2022 - 2026 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################

import math
from typing import Optional, Tuple

import torch
from torch import Tensor

from natten.libnatten import (  # type: ignore[import-untyped]
    blackwell_fmha_backward as blackwell_fmha_backward_cxx,
    blackwell_fmha_forward as blackwell_fmha_forward_cxx,
    blackwell_na1d_backward as blackwell_na1d_backward_cxx,
    blackwell_na1d_forward as blackwell_na1d_forward_cxx,
    blackwell_na2d_backward as blackwell_na2d_backward_cxx,
    blackwell_na2d_forward as blackwell_na2d_forward_cxx,
    blackwell_na3d_backward as blackwell_na3d_backward_cxx,
    blackwell_na3d_forward as blackwell_na3d_forward_cxx,
    compute_delta as compute_delta_cxx,
    fmha_backward as fmha_backward_cxx,
    fmha_forward as fmha_forward_cxx,
    hopper_fmha_backward as hopper_fmha_backward_cxx,
    hopper_fmha_forward as hopper_fmha_forward_cxx,
    hopper_na1d_backward as hopper_na1d_backward_cxx,
    hopper_na1d_forward as hopper_na1d_forward_cxx,
    hopper_na2d_backward as hopper_na2d_backward_cxx,
    hopper_na2d_forward as hopper_na2d_forward_cxx,
    hopper_na3d_backward as hopper_na3d_backward_cxx,
    hopper_na3d_forward as hopper_na3d_forward_cxx,
    na1d_backward as na1d_backward_cxx,
    na1d_forward as na1d_forward_cxx,
    na2d_backward as na2d_backward_cxx,
    na2d_forward as na2d_forward_cxx,
    na3d_backward as na3d_backward_cxx,
    na3d_forward as na3d_forward_cxx,
    reference_na1d_backward as reference_na1d_backward_cxx,
    reference_na1d_forward as reference_na1d_forward_cxx,
    reference_na2d_backward as reference_na2d_backward_cxx,
    reference_na2d_forward as reference_na2d_forward_cxx,
    reference_na3d_backward as reference_na3d_backward_cxx,
    reference_na3d_forward as reference_na3d_forward_cxx,
    token_permute_1d as token_permute_1d_cxx,
    token_permute_2d as token_permute_2d_cxx,
    token_permute_3d as token_permute_3d_cxx,
    token_unpermute_1d as token_unpermute_1d_cxx,
    token_unpermute_2d as token_unpermute_2d_cxx,
    token_unpermute_3d as token_unpermute_3d_cxx,
)
from natten.utils.environment import DISABLE_TORCH_OPS
from natten.utils.tuples import ceil_div_tuple, mul_tuple


def maybe_contiguous(x):
    return x.contiguous()


# https://github.com/Dao-AILab/flash-attention/blob/v2.7.4/flash_attn/flash_attn_interface.py#L65-L78
def disabled_register_op(
    name, fn=None, /, *, mutates_args, device_types=None, schema=None
):
    def wrap(func):
        return func

    if fn is None:
        return wrap
    return fn


def disabled_register_fake(op, fn=None, /, *, lib=None, _stacklevel=1):
    def wrap(func):
        return func

    if fn is None:
        return wrap
    return fn


if DISABLE_TORCH_OPS:
    register_op = disabled_register_op
    register_fake = disabled_register_fake
else:
    register_op = torch.library.custom_op
    register_fake = torch.library.register_fake

################################################################################
################################### FMHA ops ###################################
################################################################################


# blackwell_fmha_forward
@register_op("natten::blackwell_fmha_forward", mutates_args=(), device_types="cuda")
def blackwell_fmha_forward_torch_op(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    is_causal: bool,
    scale: float,
    q_tile_size: int,
    kv_tile_size: int,
    run_persistent_kernel: bool,
    cumulative_seqlen_Q: Optional[Tensor],
    cumulative_seqlen_KV: Optional[Tensor],
    max_seqlen_Q: int,
    max_seqlen_KV: int,
) -> Tuple[Tensor, Tensor]:
    query, key, value = [maybe_contiguous(x) for x in (query, key, value)]

    output_shape = [s for s in query.shape[:-1]] + [value.shape[-1]]

    # NOTE: always zero-init outputs when doing varlen for safety
    is_varlen = cumulative_seqlen_Q is not None
    init_fn = torch.zeros if is_varlen else torch.empty

    output = init_fn(
        output_shape, device=query.device, dtype=query.dtype
    )  # type: ignore[operator]
    logsumexp = init_fn(
        query.shape[:-1], dtype=torch.float32, device=query.device
    )  # type: ignore[operator]

    blackwell_fmha_forward_cxx(
        output,
        query,
        key,
        value,
        logsumexp,
        is_causal,
        scale,
        q_tile_size,
        kv_tile_size,
        run_persistent_kernel,
        cumulative_seqlen_Q,
        cumulative_seqlen_KV,
        max_seqlen_Q,
        max_seqlen_KV,
    )

    return output, logsumexp


@register_fake("natten::blackwell_fmha_forward")
def blackwell_fmha_forward_torch_fake_op(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    is_causal: bool,
    scale: float,
    q_tile_size: int,
    kv_tile_size: int,
    run_persistent_kernel: bool,
    cumulative_seqlen_Q: Optional[Tensor],
    cumulative_seqlen_KV: Optional[Tensor],
    max_seqlen_Q: int,
    max_seqlen_KV: int,
) -> Tuple[Tensor, Tensor]:
    query, key, value = [maybe_contiguous(x) for x in (query, key, value)]

    output_shape = [s for s in query.shape[:-1]] + [value.shape[-1]]
    output = torch.empty(output_shape, device=query.device, dtype=query.dtype)

    logsumexp = torch.empty(query.shape[:-1], dtype=torch.float32, device=query.device)
    return output, logsumexp


# blackwell_fmha_backward
@register_op(
    "natten::blackwell_fmha_backward",
    mutates_args=(),
    device_types="cuda",
)
def blackwell_fmha_backward_torch_op(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    output: Tensor,
    d_output: Tensor,
    logsumexp: Tensor,
    is_causal: bool,
    scale: float,
    q_tile_size: int,
    kv_tile_size: int,
    cumulative_seqlen_Q: Optional[Tensor],
    cumulative_seqlen_KV: Optional[Tensor],
    max_seqlen_Q: int,
    max_seqlen_KV: int,
) -> Tuple[Tensor, Tensor, Tensor]:
    query, key, value = [maybe_contiguous(x) for x in (query, key, value)]
    output, d_output, logsumexp = [
        maybe_contiguous(x) for x in (output, d_output, logsumexp)
    ]

    # NOTE: always zero-init outputs when doing varlen for safety
    is_varlen = cumulative_seqlen_Q is not None
    init_fn = torch.zeros_like if is_varlen else torch.empty_like

    d_query = init_fn(query)
    d_key = init_fn(key)
    d_value = init_fn(value)

    blackwell_fmha_backward_cxx(
        d_query,
        d_key,
        d_value,
        query,
        key,
        value,
        output,
        d_output,
        logsumexp,
        is_causal,
        scale,
        q_tile_size,
        kv_tile_size,
        cumulative_seqlen_Q,
        cumulative_seqlen_KV,
        max_seqlen_Q,
        max_seqlen_KV,
    )

    return d_query, d_key, d_value


@register_fake("natten::blackwell_fmha_backward")
def blackwell_fmha_backward_torch_fake_op(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    output: Tensor,
    d_output: Tensor,
    logsumexp: Tensor,
    is_causal: bool,
    scale: float,
    q_tile_size: int,
    kv_tile_size: int,
    cumulative_seqlen_Q: Optional[Tensor],
    cumulative_seqlen_KV: Optional[Tensor],
    max_seqlen_Q: int,
    max_seqlen_KV: int,
) -> Tuple[Tensor, Tensor, Tensor]:
    query, key, value = [maybe_contiguous(x) for x in (query, key, value)]
    output, d_output, logsumexp = [
        maybe_contiguous(x) for x in (output, d_output, logsumexp)
    ]

    d_query = torch.empty_like(query)
    d_key = torch.empty_like(key)
    d_value = torch.empty_like(value)
    return d_query, d_key, d_value


# hopper_fmha_forward
@register_op("natten::hopper_fmha_forward", mutates_args=(), device_types="cuda")
def hopper_fmha_forward_torch_op(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    scale: float,
    q_tile_size: int,
    kv_tile_size: int,
    kernel_schedule_int: int,
) -> Tuple[Tensor, Tensor]:
    query, key, value = [maybe_contiguous(x) for x in (query, key, value)]

    output_shape = [s for s in query.shape[:-1]] + [value.shape[-1]]
    output = torch.empty(output_shape, device=query.device, dtype=query.dtype)

    logsumexp = torch.empty(query.shape[:-1], dtype=torch.float32, device=query.device)

    hopper_fmha_forward_cxx(
        output,
        query,
        key,
        value,
        logsumexp,
        scale,
        q_tile_size,
        kv_tile_size,
        kernel_schedule_int,
    )

    return output, logsumexp


@register_fake("natten::hopper_fmha_forward")
def hopper_fmha_forward_torch_fake_op(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    scale: float,
    q_tile_size: int,
    kv_tile_size: int,
    kernel_schedule_int: int,
):
    query, key, value = [maybe_contiguous(x) for x in (query, key, value)]

    output_shape = [s for s in query.shape[:-1]] + [value.shape[-1]]
    output = torch.empty(output_shape, device=query.device, dtype=query.dtype)

    logsumexp = torch.empty(query.shape[:-1], dtype=torch.float32, device=query.device)

    return output, logsumexp


# hopper_fmha_backward
@register_op(
    "natten::hopper_fmha_backward",
    mutates_args=(),
    device_types="cuda",
)
def hopper_fmha_backward_torch_op(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    output: Tensor,
    d_output: Tensor,
    logsumexp: Tensor,
    scale: float,
    q_tile_size: int,
    kv_tile_size: int,
) -> Tuple[Tensor, Tensor, Tensor]:
    query, key, value = [maybe_contiguous(x) for x in (query, key, value)]
    output, d_output, logsumexp = [
        maybe_contiguous(x) for x in (output, d_output, logsumexp)
    ]

    d_query = torch.empty_like(query)
    d_key = torch.empty_like(key)
    d_value = torch.empty_like(value)

    hopper_fmha_backward_cxx(
        d_query,
        d_key,
        d_value,
        query,
        key,
        value,
        output,
        d_output,
        logsumexp,
        scale,
        q_tile_size,
        kv_tile_size,
    )

    return d_query, d_key, d_value


@register_fake("natten::hopper_fmha_backward")
def hopper_fmha_backward_torch_fake_op(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    output: Tensor,
    d_output: Tensor,
    logsumexp: Tensor,
    scale: float,
    q_tile_size: int,
    kv_tile_size: int,
):
    query, key, value = [maybe_contiguous(x) for x in (query, key, value)]
    output, d_output, logsumexp = [
        maybe_contiguous(x) for x in (output, d_output, logsumexp)
    ]

    d_query = torch.empty_like(query)
    d_key = torch.empty_like(key)
    d_value = torch.empty_like(value)

    return d_query, d_key, d_value


# fmha_forward
@register_op(
    "natten::fmha_forward",
    mutates_args=(),
    device_types="cuda",
)
def fmha_forward_torch_op(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    is_causal: bool,
    scale: float,
    q_tile_size: int,
    kv_tile_size: int,
    cumulative_seqlen_Q: Optional[Tensor],
    cumulative_seqlen_KV: Optional[Tensor],
    max_seqlen_Q: int,
    max_seqlen_KV: int,
) -> Tuple[Tensor, Tensor]:
    query, key, value = [maybe_contiguous(x) for x in (query, key, value)]

    output_shape = [s for s in query.shape[:-1]] + [value.shape[-1]]

    # NOTE: always zero-init outputs when doing varlen for safety
    is_varlen = cumulative_seqlen_Q is not None
    init_fn = torch.zeros if is_varlen else torch.empty

    output = init_fn(
        output_shape, device=query.device, dtype=query.dtype
    )  # type: ignore[operator]
    logsumexp = init_fn(
        query.shape[:-1], dtype=torch.float32, device=query.device
    )  # type: ignore[operator]

    fmha_forward_cxx(
        output,
        query,
        key,
        value,
        logsumexp,
        is_causal,
        scale,
        q_tile_size,
        kv_tile_size,
        cumulative_seqlen_Q,
        cumulative_seqlen_KV,
        max_seqlen_Q,
        max_seqlen_KV,
    )

    return output, logsumexp


@register_fake("natten::fmha_forward")
def fmha_forward_torch_fake_op(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    is_causal: bool,
    scale: float,
    q_tile_size: int,
    kv_tile_size: int,
    cumulative_seqlen_Q: Optional[Tensor],
    cumulative_seqlen_KV: Optional[Tensor],
    max_seqlen_Q: int,
    max_seqlen_KV: int,
) -> Tuple[Tensor, Tensor]:
    query, key, value = [maybe_contiguous(x) for x in (query, key, value)]

    output_shape = [s for s in query.shape[:-1]] + [value.shape[-1]]
    output = torch.empty(output_shape, device=query.device, dtype=query.dtype)

    logsumexp = torch.empty(query.shape[:-1], dtype=torch.float32, device=query.device)
    return output, logsumexp


# fmha_backward
@register_op(
    "natten::fmha_backward",
    mutates_args=(),
    device_types="cuda",
)
def fmha_backward_torch_op(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    output: Tensor,
    d_output: Tensor,
    logsumexp: Tensor,
    is_causal: bool,
    scale: float,
    q_tile_size: int,
    kv_tile_size: int,
    num_kv_splits: int,
    compute_delta_with_pt: bool,
    cumulative_seqlen_Q: Optional[Tensor],
    cumulative_seqlen_KV: Optional[Tensor],
    max_seqlen_Q: int,
    max_seqlen_KV: int,
) -> Tuple[Tensor, Tensor, Tensor]:
    query, key, value = [maybe_contiguous(x) for x in (query, key, value)]
    output, d_output, logsumexp = [
        maybe_contiguous(x) for x in (output, d_output, logsumexp)
    ]

    # NOTE: always zero-init outputs when doing varlen for safety
    is_varlen = cumulative_seqlen_Q is not None
    init_fn = torch.zeros_like if is_varlen else torch.empty_like

    d_query = init_fn(query)
    d_key = init_fn(key)
    d_value = init_fn(value)

    # torch compile sometimes passes down an incorrect num_kv_splits after recompiles.
    # we always need to silently fall back to maximum possible to avoid hitting the cpp guard.
    seqlen_kv = key.shape[1] if cumulative_seqlen_KV is None else max_seqlen_KV
    num_kv_tiles = (seqlen_kv + kv_tile_size - 1) // kv_tile_size
    assert num_kv_tiles > 0
    num_kv_splits = min(num_kv_tiles, num_kv_splits)

    fmha_backward_cxx(
        d_query,
        d_key,
        d_value,
        query,
        key,
        value,
        output,
        d_output,
        logsumexp,
        is_causal,
        scale,
        q_tile_size,
        kv_tile_size,
        num_kv_splits,
        compute_delta_with_pt,
        cumulative_seqlen_Q,
        cumulative_seqlen_KV,
        max_seqlen_Q,
        max_seqlen_KV,
    )

    return d_query, d_key, d_value


@register_fake("natten::fmha_backward")
def fmha_backward_torch_fake_op(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    output: Tensor,
    d_output: Tensor,
    logsumexp: Tensor,
    is_causal: bool,
    scale: float,
    q_tile_size: int,
    kv_tile_size: int,
    num_kv_splits: int,
    compute_delta_with_pt: bool,
    cumulative_seqlen_Q: Optional[Tensor],
    cumulative_seqlen_KV: Optional[Tensor],
    max_seqlen_Q: int,
    max_seqlen_KV: int,
) -> Tuple[Tensor, Tensor, Tensor]:
    query, key, value = [maybe_contiguous(x) for x in (query, key, value)]
    output, d_output, logsumexp = [
        maybe_contiguous(x) for x in (output, d_output, logsumexp)
    ]

    d_query = torch.empty_like(query)
    d_key = torch.empty_like(key)
    d_value = torch.empty_like(value)
    return d_query, d_key, d_value


################################################################################
################################### FNA ops  ###################################
################################################################################


def make_blackwell_fna_ops(na_dim):
    fwd_handle, bwd_handle = {
        1: (blackwell_na1d_forward_cxx, blackwell_na1d_backward_cxx),
        2: (blackwell_na2d_forward_cxx, blackwell_na2d_backward_cxx),
        3: (blackwell_na3d_forward_cxx, blackwell_na3d_backward_cxx),
    }[na_dim]

    # blackwell_na*d_forward
    @register_op(
        f"natten::blackwell_na{na_dim}d_forward",
        mutates_args=(),
        device_types="cuda",
    )
    def blackwell_fna_forward_torch_op(
        query: Tensor,
        key: Tensor,
        value: Tensor,
        kernel_size: list[int],
        stride: list[int],
        dilation: list[int],
        is_causal: list[bool],
        scale: float,
        q_shape: list[int],
        kv_shape: list[int],
        qkv_shape: list[int],
        q_tile_shape: list[int],
        kv_tile_shape: list[int],
        run_persistent_kernel: bool,
    ) -> Tuple[Tensor, Tensor]:
        query, key, value = [maybe_contiguous(x) for x in (query, key, value)]

        output_shape = [s for s in query.shape[:-1]] + [value.shape[-1]]
        output = torch.empty(output_shape, device=query.device, dtype=query.dtype)

        logsumexp = torch.empty(
            query.shape[:-1], dtype=torch.float32, device=query.device
        )

        fwd_handle(
            output,
            query,
            key,
            value,
            logsumexp,
            kernel_size,
            stride,
            dilation,
            is_causal,
            scale,
            q_shape,
            kv_shape,
            qkv_shape,
            q_tile_shape,
            kv_tile_shape,
            run_persistent_kernel,
        )

        return output, logsumexp

    @register_fake(f"natten::blackwell_na{na_dim}d_forward")
    def blackwell_fna_forward_torch_fake_op(
        query: Tensor,
        key: Tensor,
        value: Tensor,
        kernel_size: list[int],
        stride: list[int],
        dilation: list[int],
        is_causal: list[bool],
        scale: float,
        q_shape: list[int],
        kv_shape: list[int],
        qkv_shape: list[int],
        q_tile_shape: list[int],
        kv_tile_shape: list[int],
        run_persistent_kernel: bool,
    ) -> Tuple[Tensor, Tensor]:
        query, key, value = [maybe_contiguous(x) for x in (query, key, value)]

        output_shape = [s for s in query.shape[:-1]] + [value.shape[-1]]
        output = torch.empty(output_shape, device=query.device, dtype=query.dtype)

        logsumexp = torch.empty(
            query.shape[:-1], dtype=torch.float32, device=query.device
        )

        return output, logsumexp

    # blackwell_na*d_backward
    @register_op(
        f"natten::blackwell_na{na_dim}d_backward",
        mutates_args=(),
        device_types="cuda",
    )
    def blackwell_fna_backward_torch_op(
        query: Tensor,
        key: Tensor,
        value: Tensor,
        output: Tensor,
        d_output: Tensor,
        logsumexp: Tensor,
        kernel_size: list[int],
        stride: list[int],
        dilation: list[int],
        is_causal: list[bool],
        scale: float,
        q_shape: list[int],
        kv_shape: list[int],
        qkv_shape: list[int],
        q_tile_shape: list[int],
        kv_tile_shape: list[int],
    ) -> Tuple[Tensor, Tensor, Tensor]:
        query, key, value = [maybe_contiguous(x) for x in (query, key, value)]
        output, d_output, logsumexp = [
            maybe_contiguous(x) for x in (output, d_output, logsumexp)
        ]

        d_query = torch.empty_like(query)
        d_key = torch.empty_like(key)
        d_value = torch.empty_like(value)

        bwd_handle(
            d_query,
            d_key,
            d_value,
            query,
            key,
            value,
            output,
            d_output,
            logsumexp,
            kernel_size,
            stride,
            dilation,
            is_causal,
            scale,
            q_shape,
            kv_shape,
            qkv_shape,
            q_tile_shape,
            kv_tile_shape,
        )

        return d_query, d_key, d_value

    @register_fake(f"natten::blackwell_na{na_dim}d_backward")
    def blackwell_fna_backward_torch_fake_op(
        query: Tensor,
        key: Tensor,
        value: Tensor,
        output: Tensor,
        d_output: Tensor,
        logsumexp: Tensor,
        kernel_size: list[int],
        stride: list[int],
        dilation: list[int],
        is_causal: list[bool],
        scale: float,
        q_shape: list[int],
        kv_shape: list[int],
        qkv_shape: list[int],
        q_tile_shape: list[int],
        kv_tile_shape: list[int],
    ) -> Tuple[Tensor, Tensor, Tensor]:
        query, key, value = [maybe_contiguous(x) for x in (query, key, value)]
        output, d_output, logsumexp = [
            maybe_contiguous(x) for x in (output, d_output, logsumexp)
        ]

        d_query = torch.empty_like(query)
        d_key = torch.empty_like(key)
        d_value = torch.empty_like(value)

        return d_query, d_key, d_value

    return (
        blackwell_fna_forward_torch_op,
        blackwell_fna_forward_torch_fake_op,
        blackwell_fna_backward_torch_op,
        blackwell_fna_backward_torch_fake_op,
    )


def make_hopper_fna_ops(na_dim):
    fwd_handle, bwd_handle = {
        1: (hopper_na1d_forward_cxx, hopper_na1d_backward_cxx),
        2: (hopper_na2d_forward_cxx, hopper_na2d_backward_cxx),
        3: (hopper_na3d_forward_cxx, hopper_na3d_backward_cxx),
    }[na_dim]

    # hopper_na*d_forward
    @register_op(
        f"natten::hopper_na{na_dim}d_forward",
        mutates_args=(),
        device_types="cuda",
    )
    def hopper_fna_forward_torch_op(
        query: Tensor,
        key: Tensor,
        value: Tensor,
        kernel_size: list[int],
        stride: list[int],
        dilation: list[int],
        is_causal: list[bool],
        scale: float,
        q_shape: list[int],
        kv_shape: list[int],
        qkv_shape: list[int],
        q_tile_shape: list[int],
        kv_tile_shape: list[int],
        kernel_schedule_int: int,
    ) -> Tuple[Tensor, Tensor]:
        query, key, value = [maybe_contiguous(x) for x in (query, key, value)]

        output_shape = [s for s in query.shape[:-1]] + [value.shape[-1]]
        output = torch.empty(output_shape, device=query.device, dtype=query.dtype)

        logsumexp = torch.empty(
            query.shape[:-1], dtype=torch.float32, device=query.device
        )

        fwd_handle(
            output,
            query,
            key,
            value,
            logsumexp,
            kernel_size,
            stride,
            dilation,
            is_causal,
            scale,
            q_shape,
            kv_shape,
            qkv_shape,
            q_tile_shape,
            kv_tile_shape,
            kernel_schedule_int,
        )

        return output, logsumexp

    @register_fake(f"natten::hopper_na{na_dim}d_forward")
    def hopper_fna_forward_torch_fake_op(
        query: Tensor,
        key: Tensor,
        value: Tensor,
        kernel_size: list[int],
        stride: list[int],
        dilation: list[int],
        is_causal: list[bool],
        scale: float,
        q_shape: list[int],
        kv_shape: list[int],
        qkv_shape: list[int],
        q_tile_shape: list[int],
        kv_tile_shape: list[int],
        kernel_schedule_int: int,
    ) -> Tuple[Tensor, Tensor]:
        query, key, value = [maybe_contiguous(x) for x in (query, key, value)]

        output_shape = [s for s in query.shape[:-1]] + [value.shape[-1]]
        output = torch.empty(output_shape, device=query.device, dtype=query.dtype)

        logsumexp = torch.empty(
            query.shape[:-1], dtype=torch.float32, device=query.device
        )

        return output, logsumexp

    # hopper_na*d_backward
    @register_op(
        f"natten::hopper_na{na_dim}d_backward",
        mutates_args=(),
        device_types="cuda",
    )
    def hopper_fna_backward_torch_op(
        query: Tensor,
        key: Tensor,
        value: Tensor,
        output: Tensor,
        d_output: Tensor,
        logsumexp: Tensor,
        kernel_size: list[int],
        stride: list[int],
        dilation: list[int],
        is_causal: list[bool],
        scale: float,
        q_shape: list[int],
        kv_shape: list[int],
        qkv_shape: list[int],
        q_tile_shape: list[int],
        kv_tile_shape: list[int],
    ) -> Tuple[Tensor, Tensor, Tensor]:
        query, key, value = [maybe_contiguous(x) for x in (query, key, value)]
        output, d_output, logsumexp = [
            maybe_contiguous(x) for x in (output, d_output, logsumexp)
        ]

        d_query = torch.empty_like(query)
        d_key = torch.empty_like(key)
        d_value = torch.empty_like(value)

        bwd_handle(
            d_query,
            d_key,
            d_value,
            query,
            key,
            value,
            output,
            d_output,
            logsumexp,
            kernel_size,
            stride,
            dilation,
            is_causal,
            scale,
            q_shape,
            kv_shape,
            qkv_shape,
            q_tile_shape,
            kv_tile_shape,
        )

        return d_query, d_key, d_value

    @register_fake(f"natten::hopper_na{na_dim}d_backward")
    def hopper_fna_backward_torch_fake_op(
        query: Tensor,
        key: Tensor,
        value: Tensor,
        output: Tensor,
        d_output: Tensor,
        logsumexp: Tensor,
        kernel_size: list[int],
        stride: list[int],
        dilation: list[int],
        is_causal: list[bool],
        scale: float,
        q_shape: list[int],
        kv_shape: list[int],
        qkv_shape: list[int],
        q_tile_shape: list[int],
        kv_tile_shape: list[int],
    ) -> Tuple[Tensor, Tensor, Tensor]:
        query, key, value = [maybe_contiguous(x) for x in (query, key, value)]
        output, d_output, logsumexp = [
            maybe_contiguous(x) for x in (output, d_output, logsumexp)
        ]

        d_query = torch.empty_like(query)
        d_key = torch.empty_like(key)
        d_value = torch.empty_like(value)

        return d_query, d_key, d_value

    return (
        hopper_fna_forward_torch_op,
        hopper_fna_forward_torch_fake_op,
        hopper_fna_backward_torch_op,
        hopper_fna_backward_torch_fake_op,
    )


def make_fna_ops(na_dim):
    fwd_handle, bwd_handle = {
        1: (na1d_forward_cxx, na1d_backward_cxx),
        2: (na2d_forward_cxx, na2d_backward_cxx),
        3: (na3d_forward_cxx, na3d_backward_cxx),
    }[na_dim]

    # na*d_forward
    @register_op(
        f"natten::na{na_dim}d_forward",
        mutates_args=(),
        device_types="cuda",
    )
    def fna_forward_torch_op(
        query: Tensor,
        key: Tensor,
        value: Tensor,
        kernel_size: list[int],
        stride: list[int],
        dilation: list[int],
        is_causal: list[bool],
        scale: float,
        q_tile_shape: list[int],
        kv_tile_shape: list[int],
    ) -> Tuple[Tensor, Tensor]:
        query, key, value = [maybe_contiguous(x) for x in (query, key, value)]

        output_shape = [s for s in query.shape[:-1]] + [value.shape[-1]]
        output = torch.empty(output_shape, device=query.device, dtype=query.dtype)

        logsumexp = torch.empty(
            query.shape[:-1], dtype=torch.float32, device=query.device
        )

        fwd_handle(
            output,
            query,
            key,
            value,
            logsumexp,
            kernel_size,
            stride,
            dilation,
            is_causal,
            scale,
            q_tile_shape,
            kv_tile_shape,
        )

        return output, logsumexp

    @register_fake(f"natten::na{na_dim}d_forward")
    def fna_forward_torch_fake_op(
        query: Tensor,
        key: Tensor,
        value: Tensor,
        kernel_size: list[int],
        stride: list[int],
        dilation: list[int],
        is_causal: list[bool],
        scale: float,
        q_tile_shape: list[int],
        kv_tile_shape: list[int],
    ) -> Tuple[Tensor, Tensor]:
        query, key, value = [maybe_contiguous(x) for x in (query, key, value)]

        output_shape = [s for s in query.shape[:-1]] + [value.shape[-1]]
        output = torch.empty(output_shape, device=query.device, dtype=query.dtype)

        logsumexp = torch.empty(
            query.shape[:-1], dtype=torch.float32, device=query.device
        )

        return output, logsumexp

    # na*d_backward
    @register_op(
        f"natten::na{na_dim}d_backward",
        mutates_args=(),
        device_types="cuda",
    )
    def fna_backward_torch_op(
        query: Tensor,
        key: Tensor,
        value: Tensor,
        output: Tensor,
        d_output: Tensor,
        logsumexp: Tensor,
        kernel_size: list[int],
        stride: list[int],
        dilation: list[int],
        is_causal: list[bool],
        scale: float,
        q_tile_shape: list[int],
        kv_tile_shape: list[int],
        num_kv_splits: list[int],
        compute_delta_with_pt: bool,
    ) -> Tuple[Tensor, Tensor, Tensor]:
        query, key, value = [maybe_contiguous(x) for x in (query, key, value)]
        output, d_output, logsumexp = [
            maybe_contiguous(x) for x in (output, d_output, logsumexp)
        ]

        d_query = torch.empty_like(query)
        d_key = torch.empty_like(key)
        d_value = torch.empty_like(value)

        bwd_handle(
            d_query,
            d_key,
            d_value,
            query,
            key,
            value,
            output,
            d_output,
            logsumexp,
            kernel_size,
            stride,
            dilation,
            is_causal,
            scale,
            q_tile_shape,
            kv_tile_shape,
            num_kv_splits,
            compute_delta_with_pt,
        )

        return d_query, d_key, d_value

    @register_fake(f"natten::na{na_dim}d_backward")
    def fna_backward_torch_fake_op(
        query: Tensor,
        key: Tensor,
        value: Tensor,
        output: Tensor,
        d_output: Tensor,
        logsumexp: Tensor,
        kernel_size: list[int],
        stride: list[int],
        dilation: list[int],
        is_causal: list[bool],
        scale: float,
        q_tile_shape: list[int],
        kv_tile_shape: list[int],
        num_kv_splits: list[int],
        compute_delta_with_pt: bool,
    ) -> Tuple[Tensor, Tensor, Tensor]:
        query, key, value = [maybe_contiguous(x) for x in (query, key, value)]
        output, d_output, logsumexp = [
            maybe_contiguous(x) for x in (output, d_output, logsumexp)
        ]

        d_query = torch.empty_like(query)
        d_key = torch.empty_like(key)
        d_value = torch.empty_like(value)

        return d_query, d_key, d_value

    return (
        fna_forward_torch_op,
        fna_forward_torch_fake_op,
        fna_backward_torch_op,
        fna_backward_torch_fake_op,
    )


def make_reference_fna_ops(na_dim):
    fwd_handle, bwd_handle = {
        1: (reference_na1d_forward_cxx, reference_na1d_backward_cxx),
        2: (reference_na2d_forward_cxx, reference_na2d_backward_cxx),
        3: (reference_na3d_forward_cxx, reference_na3d_backward_cxx),
    }[na_dim]

    # na*d_forward
    @register_op(
        f"natten::reference_na{na_dim}d_forward",
        mutates_args=(),
        device_types="cuda",
    )
    def reference_fna_forward_torch_op(
        query: Tensor,
        key: Tensor,
        value: Tensor,
        kernel_size: list[int],
        stride: list[int],
        dilation: list[int],
        is_causal: list[bool],
        scale: float,
        qkv_shape: list[int],
        num_extra_kv: int,
    ) -> Tuple[Tensor, Tensor]:
        query, key, value = [maybe_contiguous(x) for x in (query, key, value)]

        output_shape = [s for s in query.shape[:-1]] + [value.shape[-1]]
        output = torch.empty(output_shape, device=query.device, dtype=query.dtype)

        logsumexp = torch.empty(
            query.shape[:-1], dtype=torch.float32, device=query.device
        )

        fwd_handle(
            output,
            query,
            key,
            value,
            logsumexp,
            kernel_size,
            stride,
            dilation,
            is_causal,
            scale,
            qkv_shape,
            num_extra_kv,
        )

        return output, logsumexp

    @register_fake(f"natten::reference_na{na_dim}d_forward")
    def reference_fna_forward_torch_fake_op(
        query: Tensor,
        key: Tensor,
        value: Tensor,
        kernel_size: list[int],
        stride: list[int],
        dilation: list[int],
        is_causal: list[bool],
        scale: float,
        qkv_shape: list[int],
        num_extra_kv: int,
    ) -> Tuple[Tensor, Tensor]:
        query, key, value = [maybe_contiguous(x) for x in (query, key, value)]

        output_shape = [s for s in query.shape[:-1]] + [value.shape[-1]]
        output = torch.empty(output_shape, device=query.device, dtype=query.dtype)

        logsumexp = torch.empty(
            query.shape[:-1], dtype=torch.float32, device=query.device
        )

        return output, logsumexp

    # na*d_backward
    @register_op(
        f"natten::reference_na{na_dim}d_backward",
        mutates_args=(),
        device_types="cuda",
    )
    def reference_fna_backward_torch_op(
        query: Tensor,
        key: Tensor,
        value: Tensor,
        output: Tensor,
        d_output: Tensor,
        logsumexp: Tensor,
        kernel_size: list[int],
        stride: list[int],
        dilation: list[int],
        is_causal: list[bool],
        scale: float,
        qkv_shape: list[int],
        num_extra_kv: int,
    ) -> Tuple[Tensor, Tensor, Tensor]:
        query, key, value = [maybe_contiguous(x) for x in (query, key, value)]
        output, d_output, logsumexp = [
            maybe_contiguous(x) for x in (output, d_output, logsumexp)
        ]

        d_query = torch.empty_like(query)
        d_key = torch.empty_like(key)
        d_value = torch.empty_like(value)

        bwd_handle(
            d_query,
            d_key,
            d_value,
            query,
            key,
            value,
            output,
            d_output,
            logsumexp,
            kernel_size,
            stride,
            dilation,
            is_causal,
            scale,
            qkv_shape,
            num_extra_kv,
        )

        return d_query, d_key, d_value

    @register_fake(f"natten::reference_na{na_dim}d_backward")
    def reference_fna_backward_torch_fake_op(
        query: Tensor,
        key: Tensor,
        value: Tensor,
        output: Tensor,
        d_output: Tensor,
        logsumexp: Tensor,
        kernel_size: list[int],
        stride: list[int],
        dilation: list[int],
        is_causal: list[bool],
        scale: float,
        qkv_shape: list[int],
        num_extra_kv: int,
    ) -> Tuple[Tensor, Tensor, Tensor]:
        query, key, value = [maybe_contiguous(x) for x in (query, key, value)]
        output, d_output, logsumexp = [
            maybe_contiguous(x) for x in (output, d_output, logsumexp)
        ]

        d_query = torch.empty_like(query)
        d_key = torch.empty_like(key)
        d_value = torch.empty_like(value)

        return d_query, d_key, d_value

    return (
        reference_fna_forward_torch_op,
        reference_fna_forward_torch_fake_op,
        reference_fna_backward_torch_op,
        reference_fna_backward_torch_fake_op,
    )


################################################################################
################################# TokPerm ops  #################################
################################################################################


def make_token_permute_ops(na_dim):
    permute_handle, unpermute_handle = {
        1: (token_permute_1d_cxx, token_unpermute_1d_cxx),
        2: (token_permute_2d_cxx, token_unpermute_2d_cxx),
        3: (token_permute_3d_cxx, token_unpermute_3d_cxx),
    }[na_dim]

    @register_op(
        f"natten::token_permute_{na_dim}d",
        mutates_args=(),
        device_types="cuda",
    )
    def token_permute_torch_op(
        input_tensor: Tensor,
        tile_shape: list[int],
        dilation: list[int],
        flip_tiled_dims: bool,
    ) -> Tensor:
        input_tensor = maybe_contiguous(input_tensor)

        token_layout = tuple(x for x in input_tensor.shape[1 : na_dim + 1])
        token_layout_padded = mul_tuple(
            mul_tuple(
                ceil_div_tuple(ceil_div_tuple(token_layout, tile_shape), dilation),  # type: ignore[arg-type]
                dilation,  # type: ignore[arg-type]
            ),
            tile_shape,  # type: ignore[arg-type]
        )
        output_shape = [
            input_tensor.shape[0],
            math.prod(token_layout_padded),
            input_tensor.shape[-2],
            input_tensor.shape[-1],
        ]
        output = torch.empty(
            output_shape, device=input_tensor.device, dtype=input_tensor.dtype
        )
        permute_handle(
            output,
            input_tensor,
            tile_shape,
            dilation,
            flip_tiled_dims,
        )

        # Fold dilation in batch dimension so that attention is correct.
        output = output.reshape(
            input_tensor.shape[0] * math.prod(dilation),
            -1,
            input_tensor.shape[-2],
            input_tensor.shape[-1],
        )

        return output

    @register_fake(f"natten::token_permute_{na_dim}d")
    def token_permute_torch_fake_op(
        input_tensor: Tensor,
        tile_shape: list[int],
        dilation: list[int],
        flip_tiled_dims: bool,
    ) -> Tensor:
        input_tensor = maybe_contiguous(input_tensor)

        token_layout = tuple(x for x in input_tensor.shape[1 : na_dim + 1])
        token_layout_padded = mul_tuple(
            mul_tuple(
                ceil_div_tuple(ceil_div_tuple(token_layout, tile_shape), dilation),  # type: ignore[arg-type]
                dilation,  # type: ignore[arg-type]
            ),
            tile_shape,  # type: ignore[arg-type]
        )

        # Fold dilation in batch dimension
        num_dilation_groups = math.prod(dilation)

        output_shape = [
            input_tensor.shape[0] * num_dilation_groups,
            math.prod(token_layout_padded) // num_dilation_groups,
            input_tensor.shape[-2],
            input_tensor.shape[-1],
        ]

        output = torch.empty(
            output_shape, device=input_tensor.device, dtype=input_tensor.dtype
        )

        return output

    @register_op(
        f"natten::token_unpermute_{na_dim}d",
        mutates_args=(),
        device_types="cuda",
    )
    def token_unpermute_torch_op(
        input_tensor: Tensor,
        token_layout_shape: list[int],
        tile_shape: list[int],
        dilation: list[int],
        flip_tiled_dims: bool,
    ) -> Tensor:
        input_tensor = maybe_contiguous(input_tensor)

        # Unfold dilation in batch dimension
        num_dilation_groups = math.prod(dilation)
        assert input_tensor.shape[0] % num_dilation_groups == 0
        input_tensor = input_tensor.reshape(
            input_tensor.shape[0] // num_dilation_groups,
            -1,
            input_tensor.shape[-2],
            input_tensor.shape[-1],
        )

        output_shape = [
            input_tensor.shape[0],
            *token_layout_shape,
            input_tensor.shape[-2],
            input_tensor.shape[-1],
        ]
        output = torch.empty(
            output_shape, device=input_tensor.device, dtype=input_tensor.dtype
        )
        unpermute_handle(
            output,
            input_tensor,
            tile_shape,
            dilation,
            flip_tiled_dims,
        )

        return output

    @register_fake(f"natten::token_unpermute_{na_dim}d")
    def token_unpermute_torch_fake_op(
        input_tensor: Tensor,
        token_layout_shape: list[int],
        tile_shape: list[int],
        dilation: list[int],
        flip_tiled_dims: bool,
    ) -> Tensor:
        input_tensor = maybe_contiguous(input_tensor)

        # Unfold dilation in batch dimension
        num_dilation_groups = math.prod(dilation)
        assert input_tensor.shape[0] % num_dilation_groups == 0
        input_tensor = input_tensor.reshape(
            input_tensor.shape[0] // num_dilation_groups,
            -1,
            input_tensor.shape[-2],
            input_tensor.shape[-1],
        )

        output_shape = [
            input_tensor.shape[0],
            *token_layout_shape,
            input_tensor.shape[-2],
            input_tensor.shape[-1],
        ]
        output = torch.empty(
            output_shape, device=input_tensor.device, dtype=input_tensor.dtype
        )

        return output

    return (
        token_permute_torch_op,
        token_permute_torch_fake_op,
        token_unpermute_torch_op,
        token_unpermute_torch_fake_op,
    )


(
    blackwell_na1d_forward_torch_op,
    blackwell_na1d_forward_torch_fake_op,
    blackwell_na1d_backward_torch_op,
    blackwell_na1d_backward_torch_fake_op,
) = make_blackwell_fna_ops(1)
(
    blackwell_na2d_forward_torch_op,
    blackwell_na2d_forward_torch_fake_op,
    blackwell_na2d_backward_torch_op,
    blackwell_na2d_backward_torch_fake_op,
) = make_blackwell_fna_ops(2)
(
    blackwell_na3d_forward_torch_op,
    blackwell_na3d_forward_torch_fake_op,
    blackwell_na3d_backward_torch_op,
    blackwell_na3d_backward_torch_fake_op,
) = make_blackwell_fna_ops(3)

(
    hopper_na1d_forward_torch_op,
    hopper_na1d_forward_torch_fake_op,
    hopper_na1d_backward_torch_op,
    hopper_na1d_backward_torch_fake_op,
) = make_hopper_fna_ops(1)
(
    hopper_na2d_forward_torch_op,
    hopper_na2d_forward_torch_fake_op,
    hopper_na2d_backward_torch_op,
    hopper_na2d_backward_torch_fake_op,
) = make_hopper_fna_ops(2)
(
    hopper_na3d_forward_torch_op,
    hopper_na3d_forward_torch_fake_op,
    hopper_na3d_backward_torch_op,
    hopper_na3d_backward_torch_fake_op,
) = make_hopper_fna_ops(3)

(
    na1d_forward_torch_op,
    na1d_forward_torch_fake_op,
    na1d_backward_torch_op,
    na1d_backward_torch_fake_op,
) = make_fna_ops(1)
(
    na2d_forward_torch_op,
    na2d_forward_torch_fake_op,
    na2d_backward_torch_op,
    na2d_backward_torch_fake_op,
) = make_fna_ops(2)
(
    na3d_forward_torch_op,
    na3d_forward_torch_fake_op,
    na3d_backward_torch_op,
    na3d_backward_torch_fake_op,
) = make_fna_ops(3)

(
    reference_na1d_forward_torch_op,
    reference_na1d_forward_torch_fake_op,
    reference_na1d_backward_torch_op,
    reference_na1d_backward_torch_fake_op,
) = make_reference_fna_ops(1)
(
    reference_na2d_forward_torch_op,
    reference_na2d_forward_torch_fake_op,
    reference_na2d_backward_torch_op,
    reference_na2d_backward_torch_fake_op,
) = make_reference_fna_ops(2)
(
    reference_na3d_forward_torch_op,
    reference_na3d_forward_torch_fake_op,
    reference_na3d_backward_torch_op,
    reference_na3d_backward_torch_fake_op,
) = make_reference_fna_ops(3)

(
    token_permute_1d_torch_op,
    token_permute_1d_torch_fake_op,
    token_unpermute_1d_torch_op,
    token_unpermute_1d_torch_fake_op,
) = make_token_permute_ops(1)
(
    token_permute_2d_torch_op,
    token_permute_2d_torch_fake_op,
    token_unpermute_2d_torch_op,
    token_unpermute_2d_torch_fake_op,
) = make_token_permute_ops(2)
(
    token_permute_3d_torch_op,
    token_permute_3d_torch_fake_op,
    token_unpermute_3d_torch_op,
    token_unpermute_3d_torch_fake_op,
) = make_token_permute_ops(3)


if DISABLE_TORCH_OPS:
    # Torch wrapped handles
    blackwell_fmha_forward = blackwell_fmha_forward_torch_op
    blackwell_fmha_backward = blackwell_fmha_backward_torch_op

    hopper_fmha_forward = hopper_fmha_forward_torch_op
    hopper_fmha_backward = hopper_fmha_backward_torch_op

    fmha_forward = fmha_forward_torch_op
    fmha_backward = fmha_backward_torch_op

    blackwell_na1d_forward = blackwell_na1d_forward_torch_op
    blackwell_na1d_backward = blackwell_na1d_backward_torch_op

    blackwell_na2d_forward = blackwell_na2d_forward_torch_op
    blackwell_na2d_backward = blackwell_na2d_backward_torch_op

    blackwell_na3d_forward = blackwell_na3d_forward_torch_op
    blackwell_na3d_backward = blackwell_na3d_backward_torch_op

    hopper_na1d_forward = hopper_na1d_forward_torch_op
    hopper_na1d_backward = hopper_na1d_backward_torch_op

    hopper_na2d_forward = hopper_na2d_forward_torch_op
    hopper_na2d_backward = hopper_na2d_backward_torch_op

    hopper_na3d_forward = hopper_na3d_forward_torch_op
    hopper_na3d_backward = hopper_na3d_backward_torch_op

    na1d_forward = na1d_forward_torch_op
    na1d_backward = na1d_backward_torch_op

    na2d_forward = na2d_forward_torch_op
    na2d_backward = na2d_backward_torch_op

    na3d_forward = na3d_forward_torch_op
    na3d_backward = na3d_backward_torch_op

    reference_na1d_forward = reference_na1d_forward_torch_op
    reference_na1d_backward = reference_na1d_backward_torch_op

    reference_na2d_forward = reference_na2d_forward_torch_op
    reference_na2d_backward = reference_na2d_backward_torch_op

    reference_na3d_forward = reference_na3d_forward_torch_op
    reference_na3d_backward = reference_na3d_backward_torch_op

    token_permute_1d = token_permute_1d_torch_op
    token_permute_2d = token_permute_2d_torch_op
    token_permute_3d = token_permute_3d_torch_op

    token_unpermute_1d = token_unpermute_1d_torch_op
    token_unpermute_2d = token_unpermute_2d_torch_op
    token_unpermute_3d = token_unpermute_3d_torch_op

else:
    # Torch wrapped handles
    blackwell_fmha_forward = torch.ops.natten.blackwell_fmha_forward
    blackwell_fmha_backward = torch.ops.natten.blackwell_fmha_backward

    hopper_fmha_forward = torch.ops.natten.hopper_fmha_forward
    hopper_fmha_backward = torch.ops.natten.hopper_fmha_backward

    fmha_forward = torch.ops.natten.fmha_forward
    fmha_backward = torch.ops.natten.fmha_backward

    blackwell_na1d_forward = torch.ops.natten.blackwell_na1d_forward
    blackwell_na1d_backward = torch.ops.natten.blackwell_na1d_backward

    blackwell_na2d_forward = torch.ops.natten.blackwell_na2d_forward
    blackwell_na2d_backward = torch.ops.natten.blackwell_na2d_backward

    blackwell_na3d_forward = torch.ops.natten.blackwell_na3d_forward
    blackwell_na3d_backward = torch.ops.natten.blackwell_na3d_backward

    hopper_na1d_forward = torch.ops.natten.hopper_na1d_forward
    hopper_na1d_backward = torch.ops.natten.hopper_na1d_backward

    hopper_na2d_forward = torch.ops.natten.hopper_na2d_forward
    hopper_na2d_backward = torch.ops.natten.hopper_na2d_backward

    hopper_na3d_forward = torch.ops.natten.hopper_na3d_forward
    hopper_na3d_backward = torch.ops.natten.hopper_na3d_backward

    na1d_forward = torch.ops.natten.na1d_forward
    na1d_backward = torch.ops.natten.na1d_backward

    na2d_forward = torch.ops.natten.na2d_forward
    na2d_backward = torch.ops.natten.na2d_backward

    na3d_forward = torch.ops.natten.na3d_forward
    na3d_backward = torch.ops.natten.na3d_backward

    reference_na1d_forward = torch.ops.natten.reference_na1d_forward
    reference_na1d_backward = torch.ops.natten.reference_na1d_backward

    reference_na2d_forward = torch.ops.natten.reference_na2d_forward
    reference_na2d_backward = torch.ops.natten.reference_na2d_backward

    reference_na3d_forward = torch.ops.natten.reference_na3d_forward
    reference_na3d_backward = torch.ops.natten.reference_na3d_backward

    token_permute_1d = torch.ops.natten.token_permute_1d
    token_permute_2d = torch.ops.natten.token_permute_2d
    token_permute_3d = torch.ops.natten.token_permute_3d

    token_unpermute_1d = torch.ops.natten.token_unpermute_1d
    token_unpermute_2d = torch.ops.natten.token_unpermute_2d
    token_unpermute_3d = torch.ops.natten.token_unpermute_3d

# This is only used in unit tests, and not even auto-diffable
compute_delta = compute_delta_cxx


__all__ = [
    "blackwell_fmha_backward",
    "blackwell_fmha_forward",
    "blackwell_na1d_backward",
    "blackwell_na1d_forward",
    "blackwell_na2d_backward",
    "blackwell_na2d_forward",
    "blackwell_na3d_backward",
    "blackwell_na3d_forward",
    "compute_delta",
    "fmha_backward",
    "fmha_forward",
    "hopper_fmha_backward",
    "hopper_fmha_forward",
    "hopper_na1d_backward",
    "hopper_na1d_forward",
    "hopper_na2d_backward",
    "hopper_na2d_forward",
    "hopper_na3d_backward",
    "hopper_na3d_forward",
    "na1d_backward",
    "na1d_forward",
    "na2d_backward",
    "na2d_forward",
    "na3d_backward",
    "na3d_forward",
    "reference_na1d_backward",
    "reference_na1d_forward",
    "reference_na2d_backward",
    "reference_na2d_forward",
    "reference_na3d_backward",
    "reference_na3d_forward",
    "token_permute_1d",
    "token_permute_2d",
    "token_permute_3d",
    "token_unpermute_1d",
    "token_unpermute_2d",
    "token_unpermute_3d",
]



#################################################################################################
# Copyright (c) 2022 - 2026 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################

from typing import Optional, Tuple, Union

import torch  # noqa: F401
from torch import Tensor

from natten.types import NoneType
from natten.utils.environment import is_torch_compiling


def generate_varlen_parameters(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    seqlens_Q: Optional[Tensor] = None,
    seqlens_KV: Optional[Tensor] = None,
) -> Union[
    Tuple[NoneType, NoneType, int, int],
    Tuple[Tensor, Tensor, int, int],
]:
    # NOTE: max_seqlen_{Q,KV} require a device-host sync, since they're expected to be ints (with
    # which we launch the varlen kernel) and not device tensors.
    # .item() introduces control flow and breaks the graph.
    # It is also inefficient to repeat this per-op, and mostly there for convenience.
    # generate_varlen_parameters should ideally always be called by the user ahead of model
    # forward / backward.
    if is_torch_compiling():
        raise RuntimeError(
            "Running 'generate_varlen_parameters' in a torch-compiled region is disallowed as it "
            "results in graph breaks. Please consider calling ahead of time and pass "
            "'cumulative_seqlen_{Q,KV}' and 'max_seqlen_{Q,KV}' instead of 'seqlens_{Q,KV}' to "
            "'attention'. "
        )

    if query.shape[0] != key.shape[0] or query.shape[0] != value.shape[0]:
        raise ValueError(
            "Q, K, and V must match in batch size, got "
            f"{query.shape[0]=}, {key.shape[0]=}, {value.shape[0]=}."
        )

    if (seqlens_Q is None) ^ (seqlens_KV is None):
        raise ValueError(
            "Variable length Attention requires both of seqlens_Q and seqlens_KV to be set, got "
            f"{seqlens_Q=}, {seqlens_KV=}."
        )

    if seqlens_Q is None and seqlens_KV is None:
        # Not varlen
        return None, None, 0, 0

    assert seqlens_Q is not None
    assert seqlens_KV is not None

    if not isinstance(seqlens_Q, Tensor) or not isinstance(seqlens_KV, Tensor):
        raise ValueError("seqlens_Q and seqlens_KV must both be tensors.")

    if seqlens_Q.device != query.device or seqlens_KV.device != query.device:
        raise ValueError(
            "seqlens_Q and seqlens_KV must be on the same device as QKV, but "
            f"{seqlens_Q.device=}, {seqlens_KV.device=}, {query.device=}."
        )

    if seqlens_Q.dtype != torch.int32 or seqlens_KV.dtype != torch.int32:
        raise ValueError(
            "seqlens_Q and seqlens_KV must both be torch.int32 tensors, got "
            f"{seqlens_Q.dtype=}, {seqlens_KV.dtype=}."
        )

    if seqlens_Q.dim() != 1 or seqlens_KV.dim() != 1:
        raise ValueError(
            "seqlens_Q and seqlens_KV must both be 1-D tensors, got "
            f"{seqlens_Q.dim()=}, {seqlens_KV.dim()=}."
        )

    if seqlens_Q.shape[0] != seqlens_KV.shape[0]:
        raise ValueError(
            "seqlens_Q and seqlens_KV must match in size, got "
            f"{seqlens_Q.shape=}, {seqlens_KV.shape=}."
        )

    if seqlens_Q.shape[0] < 1:
        raise ValueError(
            "seqlens_Q and seqlens_KV must contain at least one element, got "
            f"{seqlens_Q.shape=}, {seqlens_KV.shape=}."
        )

    if query.shape[0] != 1:
        raise ValueError(
            "Variable length attention only supports sequence-packed memory layout "
            f"(batch = 1), got {query.shape[0]=}."
        )

    assert seqlens_Q.dim() == seqlens_KV.dim() == 1
    assert seqlens_Q.shape[0] == seqlens_KV.shape[0] >= 1
    assert seqlens_Q.dtype == seqlens_KV.dtype == torch.int32

    max_seqlen_Q = seqlens_Q.max().item()  # type: ignore
    max_seqlen_KV = seqlens_KV.max().item()  # type: ignore

    # NOTE: we have to prepend with 0 manually :(
    z = torch.tensor([0], dtype=torch.int32, device=seqlens_Q.device)
    cumulative_seqlen_Q = torch.cat([z, seqlens_Q.cumsum(0).to(torch.int32)], dim=0)
    cumulative_seqlen_KV = torch.cat([z, seqlens_KV.cumsum(0).to(torch.int32)], dim=0)

    assert isinstance(max_seqlen_Q, int)
    assert isinstance(max_seqlen_KV, int)

    return (
        cumulative_seqlen_Q,
        cumulative_seqlen_KV,
        max_seqlen_Q,
        max_seqlen_KV,
    )



#################################################################################################
# Copyright (c) 2022 - 2026 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################

import torch  # noqa: F401

try:
    from natten import libnatten  # noqa: F401

    HAS_LIBNATTEN = True

    from natten._libnatten.torch_wrappers import (
        blackwell_fmha_backward,
        blackwell_fmha_forward,
        blackwell_na1d_backward,
        blackwell_na1d_forward,
        blackwell_na2d_backward,
        blackwell_na2d_forward,
        blackwell_na3d_backward,
        blackwell_na3d_forward,
        compute_delta,
        fmha_backward,
        fmha_forward,
        hopper_fmha_backward,
        hopper_fmha_forward,
        hopper_na1d_backward,
        hopper_na1d_forward,
        hopper_na2d_backward,
        hopper_na2d_forward,
        hopper_na3d_backward,
        hopper_na3d_forward,
        na1d_backward,
        na1d_forward,
        na2d_backward,
        na2d_forward,
        na3d_backward,
        na3d_forward,
        reference_na1d_backward,
        reference_na1d_forward,
        reference_na2d_backward,
        reference_na2d_forward,
        reference_na3d_backward,
        reference_na3d_forward,
        token_permute_1d,
        token_permute_2d,
        token_permute_3d,
        token_unpermute_1d,
        token_unpermute_2d,
        token_unpermute_3d,
    )

except ImportError:
    HAS_LIBNATTEN = False

    from natten._libnatten.stubs import (
        blackwell_fmha_backward,
        blackwell_fmha_forward,
        blackwell_na1d_backward,
        blackwell_na1d_forward,
        blackwell_na2d_backward,
        blackwell_na2d_forward,
        blackwell_na3d_backward,
        blackwell_na3d_forward,
        compute_delta,
        fmha_backward,
        fmha_forward,
        hopper_fmha_backward,
        hopper_fmha_forward,
        hopper_na1d_backward,
        hopper_na1d_forward,
        hopper_na2d_backward,
        hopper_na2d_forward,
        hopper_na3d_backward,
        hopper_na3d_forward,
        na1d_backward,
        na1d_forward,
        na2d_backward,
        na2d_forward,
        na3d_backward,
        na3d_forward,
        reference_na1d_backward,
        reference_na1d_forward,
        reference_na2d_backward,
        reference_na2d_forward,
        reference_na3d_backward,
        reference_na3d_forward,
        token_permute_1d,
        token_permute_2d,
        token_permute_3d,
        token_unpermute_1d,
        token_unpermute_2d,
        token_unpermute_3d,
    )


__all__ = [
    "blackwell_fmha_backward",
    "blackwell_fmha_forward",
    "blackwell_na1d_backward",
    "blackwell_na1d_forward",
    "blackwell_na2d_backward",
    "blackwell_na2d_forward",
    "blackwell_na3d_backward",
    "blackwell_na3d_forward",
    "compute_delta",
    "fmha_backward",
    "fmha_forward",
    "hopper_fmha_backward",
    "hopper_fmha_forward",
    "hopper_na1d_backward",
    "hopper_na1d_forward",
    "hopper_na2d_backward",
    "hopper_na2d_forward",
    "hopper_na3d_backward",
    "hopper_na3d_forward",
    "na1d_backward",
    "na1d_forward",
    "na2d_backward",
    "na2d_forward",
    "na3d_backward",
    "na3d_forward",
    "reference_na1d_backward",
    "reference_na1d_forward",
    "reference_na2d_backward",
    "reference_na2d_forward",
    "reference_na3d_backward",
    "reference_na3d_forward",
    "token_permute_1d",
    "token_permute_2d",
    "token_permute_3d",
    "token_unpermute_1d",
    "token_unpermute_2d",
    "token_unpermute_3d",
]



#################################################################################################
# Copyright (c) 2022 - 2026 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################

import os

import torch

from natten.utils.device import get_device_cc


def parse_env_flag(env_var: str, default: bool) -> bool:
    default_str = "1" if default else "0"
    out_str = os.getenv(env_var, default_str)
    if out_str.strip() == "":
        return default
    if out_str == "0":
        return False
    if out_str == "1":
        return True
    return default


def parse_env_int(env_var: str, default: int) -> int:
    out_str = os.getenv(env_var, str(default))
    if out_str.strip() == "":
        return default
    try:
        return int(out_str)
    except ValueError:
        return default


def parse_env_str(env_var: str, default: str) -> str:
    return os.getenv(env_var, str(default))


_IS_CUDA_AVAILABLE = torch.cuda.is_available()

_TORCH_VERSION = [int(x) for x in torch.__version__.split(".")[:2]]

# NOTE: Triton does not recognize SM103 yet.
# TODO: remove the extra condition once Triton adds SM103 support.
_IS_TORCH_COMPILE_SUPPORTED = (
    _TORCH_VERSION >= [2, 6] and get_device_cc() >= 70 and get_device_cc() not in [103]
)

# Guard registering libnatten APIs as torch ops
# In case any unusual bugs from torch compile come up again
DISABLE_TORCH_OPS = not _IS_TORCH_COMPILE_SUPPORTED or parse_env_flag(
    "NATTEN_DISABLE_TORCH_OPS", False
)


# Controls all regions guarded against torch compile
# Logs, and certain assertions cause graph breaks.
def is_torch_compiling() -> bool:
    try:
        return torch.compiler.is_compiling()
    except:
        # Assume too old to support torch compile
        return False



#################################################################################################
# Copyright (c) 2022 - 2026 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################


def libnatten_import_error(*args, **kwargs):
    raise ImportError(
        "Failed to import libnatten. "
        "Did you build / download NATTEN with the CUDA backend (libnatten)? "
        "Please make sure you built NATTEN correctly, or refer to "
        "https://shi-labs.com/natten for more information."
    )


# Reference kernels
reference_na1d_forward = libnatten_import_error
reference_na2d_forward = libnatten_import_error
reference_na3d_forward = libnatten_import_error
reference_na1d_backward = libnatten_import_error
reference_na2d_backward = libnatten_import_error
reference_na3d_backward = libnatten_import_error

# CUTLASS 2.X kernels
fmha_forward = libnatten_import_error
fmha_backward = libnatten_import_error

## SM50/SM70/SM75/SM80 - Original FNA
na1d_forward = libnatten_import_error
na2d_forward = libnatten_import_error
na3d_forward = libnatten_import_error
na1d_backward = libnatten_import_error
na2d_backward = libnatten_import_error
na3d_backward = libnatten_import_error

# CUTLASS 3.X kernels
## SM90 - Hopper FMHA
hopper_fmha_forward = libnatten_import_error
hopper_fmha_backward = libnatten_import_error

## SM90 - Hopper FNA
hopper_na1d_forward = libnatten_import_error
hopper_na2d_forward = libnatten_import_error
hopper_na3d_forward = libnatten_import_error
hopper_na1d_backward = libnatten_import_error
hopper_na2d_backward = libnatten_import_error
hopper_na3d_backward = libnatten_import_error

## SM100 - Blackwell FMHA
blackwell_fmha_forward = libnatten_import_error
blackwell_fmha_backward = libnatten_import_error

## SM100 - Blackwell FNA
blackwell_na1d_forward = libnatten_import_error
blackwell_na2d_forward = libnatten_import_error
blackwell_na3d_forward = libnatten_import_error
blackwell_na1d_backward = libnatten_import_error
blackwell_na2d_backward = libnatten_import_error
blackwell_na3d_backward = libnatten_import_error

# Token permute kernels
token_permute_1d = libnatten_import_error
token_permute_2d = libnatten_import_error
token_permute_3d = libnatten_import_error
token_unpermute_1d = libnatten_import_error
token_unpermute_2d = libnatten_import_error
token_unpermute_3d = libnatten_import_error

# Misc kernels
compute_delta = libnatten_import_error



#################################################################################################
# Copyright (c) 2022 - 2026 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################

import functools
from typing import Tuple

import torch
from torch import Tensor
from torch.amp import custom_bwd, custom_fwd
from torch.autograd import Function

amp_fwd = functools.partial(custom_fwd, device_type="cuda")
amp_bwd = functools.partial(custom_bwd, device_type="cuda")

from natten._libnatten import (
    HAS_LIBNATTEN,
    token_permute_1d,
    token_permute_2d,
    token_permute_3d,
    token_unpermute_1d,
    token_unpermute_2d,
    token_unpermute_3d,
)
from natten.types import DimensionType, NoneType
from natten.utils import log
from natten.utils.device import get_device_cc, is_cuda

logger = log.get_logger(__name__)


def can_run_cutlass_tokperm(tensor: Tensor) -> bool:
    if not HAS_LIBNATTEN:
        logger.debug(
            "Can't use libnatten TokPerm kernels, because libnatten is not available."
        )
        return False

    if not is_cuda(tensor.device):
        logger.debug(
            "Can't use libnatten TokPerm kernels, because input is not a CUDA tensor."
        )
        return False

    is_fp8_allowed = get_device_cc(tensor.device) in [100, 103]
    if tensor.dtype not in [
        torch.float32,
        torch.float16,
        torch.bfloat16,
        torch.float16,
    ] and (
        is_fp8_allowed and tensor.dtype not in [torch.float8_e5m2, torch.float8_e4m3fn]
    ):
        logger.debug(
            f"Can't use libnatten TokPerm kernels; unexpected dtype {tensor.dtype}."
        )
        return False

    return True


PERMUTE_OPS = {1: token_permute_1d, 2: token_permute_2d, 3: token_permute_3d}
UNPERMUTE_OPS = {1: token_unpermute_1d, 2: token_unpermute_2d, 3: token_unpermute_3d}


def make_cutlass_token_permute_autograd_fn(na_dim):
    assert na_dim in [1, 2, 3]

    class CutlassTokenPermuteAutogradFn(Function):
        @staticmethod
        @amp_fwd
        def forward(
            ctx,
            tensor: Tensor,
            tile_shape: DimensionType,
            dilation: DimensionType,
            flip_tiled_dims: bool,
        ) -> Tensor:

            output = PERMUTE_OPS[na_dim](
                tensor,
                tile_shape=tile_shape,
                dilation=dilation,
                flip_tiled_dims=flip_tiled_dims,
            )

            ctx.tile_shape = tile_shape
            ctx.dilation = dilation
            ctx.flip_tiled_dims = flip_tiled_dims
            ctx.token_layout = tuple(x for x in tensor.shape[1 : na_dim + 1])
            assert len(ctx.token_layout) == na_dim

            return output

        @staticmethod
        @amp_bwd
        def backward(ctx, d_output: Tensor) -> Tuple[
            Tensor,
            NoneType,
            NoneType,
            NoneType,
        ]:

            d_output_unpermuted = UNPERMUTE_OPS[na_dim](
                d_output,
                token_layout_shape=ctx.token_layout,
                tile_shape=ctx.tile_shape,
                dilation=ctx.dilation,
                flip_tiled_dims=ctx.flip_tiled_dims,
            )

            return (
                d_output_unpermuted,
                None,
                None,
                None,
            )

    return CutlassTokenPermuteAutogradFn


def make_cutlass_token_unpermute_autograd_fn(na_dim):
    assert na_dim in [1, 2, 3]

    class CutlassTokenUnPermuteAutogradFn(Function):
        @staticmethod
        @amp_fwd
        def forward(
            ctx,
            tensor: Tensor,
            token_layout: DimensionType,
            tile_shape: DimensionType,
            dilation: DimensionType,
            flip_tiled_dims: bool,
        ) -> Tensor:

            output = UNPERMUTE_OPS[na_dim](
                tensor,
                token_layout_shape=token_layout,
                tile_shape=tile_shape,
                dilation=dilation,
                flip_tiled_dims=flip_tiled_dims,
            )

            ctx.tile_shape = tile_shape
            ctx.dilation = dilation
            ctx.flip_tiled_dims = flip_tiled_dims

            return output

        @staticmethod
        @amp_bwd
        def backward(ctx, d_output: Tensor) -> Tuple[
            Tensor,
            NoneType,
            NoneType,
            NoneType,
            NoneType,
        ]:

            d_output_permuted = PERMUTE_OPS[na_dim](
                d_output,
                tile_shape=ctx.tile_shape,
                dilation=ctx.dilation,
                flip_tiled_dims=ctx.flip_tiled_dims,
            )

            return (
                d_output_permuted,
                None,
                None,
                None,
                None,
            )

    return CutlassTokenUnPermuteAutogradFn


CutlassTokenPermute1DAutogradFn = make_cutlass_token_permute_autograd_fn(1)
CutlassTokenPermute2DAutogradFn = make_cutlass_token_permute_autograd_fn(2)
CutlassTokenPermute3DAutogradFn = make_cutlass_token_permute_autograd_fn(3)

CutlassTokenUnPermute1DAutogradFn = make_cutlass_token_unpermute_autograd_fn(1)
CutlassTokenUnPermute2DAutogradFn = make_cutlass_token_unpermute_autograd_fn(2)
CutlassTokenUnPermute3DAutogradFn = make_cutlass_token_unpermute_autograd_fn(3)

CutlassTokenPermuteAutogradFns = {
    1: CutlassTokenPermute1DAutogradFn,
    2: CutlassTokenPermute2DAutogradFn,
    3: CutlassTokenPermute3DAutogradFn,
}

CutlassTokenUnPermuteAutogradFns = {
    1: CutlassTokenUnPermute1DAutogradFn,
    2: CutlassTokenUnPermute2DAutogradFn,
    3: CutlassTokenUnPermute3DAutogradFn,
}


def token_permute_cutlass(
    tensor: Tensor,
    tile_shape: DimensionType,
    dilation: DimensionType,
    flip_tiled_dims,
) -> Tensor:
    if tensor.dim() not in [4, 5, 6]:
        raise ValueError(
            "Expected 4D, 5D, or 6D tensor (corresponding to NA1D, 2D, 3D), "
            f"got {tensor.dim()}D input."
        )

    na_dim = tensor.dim() - 3
    assert na_dim in [1, 2, 3]

    if len(tile_shape) != na_dim:
        raise ValueError(
            f"Expected {na_dim}D tiler for NA{na_dim}D, " f"got {tile_shape=}."
        )

    if not can_run_cutlass_tokperm(tensor):
        raise NotImplementedError(
            "Use case is not compatible with CUTLASS Token Permute."
        )

    tensor = tensor.contiguous()
    output = CutlassTokenPermuteAutogradFns[na_dim].apply(
        tensor,
        tile_shape,
        dilation,
        flip_tiled_dims,
    )

    return output


def token_unpermute_cutlass(
    tensor: Tensor,
    token_layout_shape: DimensionType,
    tile_shape: DimensionType,
    dilation: DimensionType,
    flip_tiled_dims: bool,
) -> Tensor:
    if tensor.dim() != 4:
        raise ValueError(f"Expected flattened 4D tensor, got {tensor.dim()}D input.")

    na_dim = len(token_layout_shape)

    if len(tile_shape) != na_dim:
        raise ValueError(
            f"Expected {na_dim}D tiler for NA{na_dim}D, " f"got {tile_shape=}."
        )

    if not can_run_cutlass_tokperm(tensor):
        raise NotImplementedError(
            "Use case is not compatible with CUTLASS Token UnPermute."
        )

    tensor = tensor.contiguous()
    output = CutlassTokenUnPermuteAutogradFns[na_dim].apply(
        tensor,
        token_layout_shape,
        tile_shape,
        dilation,
        flip_tiled_dims,
    )

    return output



#################################################################################################
# Copyright (c) 2022 - 2026 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################

from natten._libnatten import HAS_LIBNATTEN  # noqa: F401
from natten.utils.environment import (
    _IS_CUDA_AVAILABLE,
    _IS_TORCH_COMPILE_SUPPORTED,
    _TORCH_VERSION,
    parse_env_flag,
    parse_env_int,
    parse_env_str,
)


# Default tokperm implementation; choices:
# NATTEN_TOKPERM_DEFAULT_IMPL="cutlass"
# NATTEN_TOKPERM_DEFAULT_IMPL="torch"
USE_TORCH_IMPL_DEFAULT = (
    parse_env_str("NATTEN_TOKPERM_DEFAULT_IMPL", "cutlass") == "torch"
)

# Unit tests
_RUN_EXTENDED_TESTS = parse_env_flag("NATTEN_RUN_EXTENDED_TESTS", False)
_RUN_FLEX_TESTS = parse_env_flag("NATTEN_RUN_FLEX_TESTS", True)
_NUM_RAND_SWEEP_TESTS = parse_env_int("NATTEN_RAND_SWEEP_TESTS", 1000)

# Profiler
DISABLE_TQDM = parse_env_flag("NATTEN_DISABLE_TQDM", False)


__all__ = [
    "HAS_LIBNATTEN",
    "_IS_CUDA_AVAILABLE",
    "_IS_TORCH_COMPILE_SUPPORTED",
    "DISABLE_TQDM",
    "_RUN_FLEX_TESTS",
    "_RUN_FLEX_TESTS",
    "_NUM_RAND_SWEEP_TESTS",
    "_TORCH_VERSION",
]



# Copyright (c) 2022 - 2026 Ali Hassani.
#
# This script is intended to emit fused kernel instantiations into
# a variable number of source files, generate appropriate headers
# and a single dispatcher interface, which will be used by the
# NATTEN API to call the kernels.
#
# NOTE: these scripts are heavily under-documented, and
# overly-repetitive, and will be replaced in future PRs.
# Please use it with caution.

import argparse
import os
from typing import List, Tuple


DEFAULT_OUTPUT_DIR = "csrc/"


SUPPORTED_CONFIGS_BACKWARD = {
    16: {
        32: [
            (64, 128, 32),
            (128, 128, 32),
        ],
        64: [
            (64, 128, 64),
            (128, 128, 64),
        ],
        128: [
            (64, 128, 128),
        ],
    },
}


KERNEL_DECL_TEMPLATE = """
void {kernel_name}(
      void* ptr_Q,
      void* ptr_K,
      void* ptr_V,
      void* ptr_O,
      void* ptr_LSE,
      void* ptr_dQ,
      void* ptr_dK,
      void* ptr_dV,
      void* ptr_dO,
      int batch_size,
      int seqlen_q,
      int seqlen_k,
      int seqlen_lse,
      int heads,
      int dim,
      int device_id,
      float attn_scale,
      cudaStream_t stream,
      at::TensorOptions tensor_options);
"""


KERNEL_IMPL_TEMPLATE = """
void {kernel_name}(
      void* ptr_Q,
      void* ptr_K,
      void* ptr_V,
      void* ptr_O,
      void* ptr_LSE,
      void* ptr_dQ,
      void* ptr_dK,
      void* ptr_dV,
      void* ptr_dO,
      int batch_size,
      int seqlen_q,
      int seqlen_k,
      int seqlen_lse,
      int heads,
      int dim,
      int device_id,
      float attn_scale,
      cudaStream_t stream,
      at::TensorOptions tensor_options) {{

  using GemmShape = {GEMMShape};
  using Kernel = natten::cuda::fmha_hopper::KernelBackward<
    {dtype}, GemmShape, false>;
  using KernelWithResidualMask = natten::cuda::fmha_hopper::KernelBackward<
    {dtype}, GemmShape, true>;

  bool no_mask_required = seqlen_q % get<0>(GemmShape{{}}) == 0 && seqlen_k % get<1>(GemmShape{{}}) == 0;
  if (no_mask_required) {{
    Kernel kernel;
    auto args = kernel.initialize(
        ptr_Q,
        ptr_K,
        ptr_V,
        ptr_O,
        ptr_LSE,
        ptr_dQ,
        ptr_dK,
        ptr_dV,
        ptr_dO,
        batch_size,
        seqlen_q,
        seqlen_k,
        seqlen_lse,
        heads,
        dim,
        device_id,
        attn_scale);

    auto bytes = static_cast<int64_t>(kernel.get_workspace_size(args));
    auto workspace = at::empty({{bytes}}, tensor_options.dtype(at::ScalarType::Byte));
    auto workspace_ptr = static_cast<void*>(workspace.data_ptr());
    kernel.run(args, workspace_ptr, stream);
  }} else {{
    KernelWithResidualMask kernel;
    auto args = kernel.initialize(
        ptr_Q,
        ptr_K,
        ptr_V,
        ptr_O,
        ptr_LSE,
        ptr_dQ,
        ptr_dK,
        ptr_dV,
        ptr_dO,
        batch_size,
        seqlen_q,
        seqlen_k,
        seqlen_lse,
        heads,
        dim,
        device_id,
        attn_scale);

    auto bytes = static_cast<int64_t>(kernel.get_workspace_size(args));
    auto workspace = at::empty({{bytes}}, tensor_options.dtype(at::ScalarType::Byte));
    auto workspace_ptr = static_cast<void*>(workspace.data_ptr());
    kernel.run(args, workspace_ptr, stream);
  }}
}}
"""


class DataType:
    def __init__(self, name, short_name, torch_name, bits):
        self.name = name
        self.bits = bits
        self.short_name = short_name
        self.torch_name = torch_name


Half = DataType("cutlass::half_t", "float16", "torch::kFloat16", 16)
BFloat = DataType("cutlass::bfloat16_t", "bfloat16", "torch::kBFloat16", 16)


def iterable_to_static_cute_tuple(shape_in) -> str:
    shape = ", ".join([f"cute::Int<{x}>" for x in shape_in])
    return f"cute::tuple<{shape}>"


class HopperFmhaInstance:
    def __init__(
        self,
        dtype: DataType,
        gemm_shape: Tuple[int, int, int],
    ):
        assert len(gemm_shape) == 3
        self.gemm_shape = gemm_shape
        self.dtype = dtype
        self.max_head_dim = gemm_shape[2]

    def get_gemm_shape_cute(self) -> str:
        return iterable_to_static_cute_tuple(self.gemm_shape)

    def get_name(self) -> str:
        name = "hopper_fmha_backward"
        name += f"_{self.dtype.short_name}"
        name += "_" + "x".join([str(x) for x in self.gemm_shape])
        return name

    def get_decl(self) -> str:
        return KERNEL_DECL_TEMPLATE.format(
            kernel_name=self.get_name(),
        )

    def get_impl(self) -> str:
        return KERNEL_IMPL_TEMPLATE.format(
            kernel_name=self.get_name(),
            GEMMShape=self.get_gemm_shape_cute(),
            dtype=self.dtype.name,
        )


def write_combined_source_file(path, filename, headers, kernels):
    source_head = []
    source_head += ["#ifdef NATTEN_WITH_CUTLASS\n"]
    source_head += ["#ifdef NATTEN_WITH_HOPPER_FNA\n"]

    source_head += ["#include <cuda_runtime.h>\n"]
    source_head += ["#include <iostream>\n"]

    source_head += ["#include <ATen/ATen.h>\n"]
    source_head += ["#include <ATen/cuda/CUDAContext.h>\n"]
    source_head += ["#include <c10/cuda/CUDAGuard.h>\n"]
    source_head += ["#include <c10/cuda/CUDAStream.h>\n"]
    source_head += ["#include <torch/extension.h>\n"]

    source_head += ["#include <natten/natten.h>\n"]
    source_head += ["#include <natten/helpers.h>\n"]

    source_head += ["#include <natten/cuda/fmha_hopper/fmha_backward.cuh>\n"]

    for header in headers:
        source_head += [f"#include <{header}>\n"]

    source_head += ["namespace natten { \n"]
    source_head += ["namespace cuda { \n"]
    source_head += ["namespace fmha_hopper { \n\n"]

    source_head = "".join(source_head)

    source_body = []
    for kernel in kernels:
        source_body += "\n\n" + kernel.get_impl() + "\n\n"
    source_body = "".join(source_body)

    source_foot = "".join(
        [
            "} // namespace fmha_hopper \n",
            "} // namespace cuda \n",
            "} // namespace natten \n",
            "#endif \n",
            "#endif \n",
            "\n",
        ]
    )
    filename = f"{path}/{filename}"
    with open(filename, "w") as f:
        f.write(source_head)
        f.write(source_body)
        f.write(source_foot)


class DTypeDispatcher:
    def __init__(self):
        self.dtypes = []
        self.name = "DISPATCH_HOPPER_FMHA_BACKWARD"

    def append(self, dtype: DataType):
        self.dtypes.append(dtype)

    def get_dispatcher(self):
        dispatcher_str = ""
        dispatcher_str += (
            f"#define {self.name}(dtype, dim, q_tile_size, kv_tile_size, ...) \\\n"
        )
        dispatcher_str += "  [&] { \\\n"
        for i, dtype in enumerate(self.dtypes):
            dispatcher_str += "    "
            if i > 0:
                dispatcher_str += "else "
            dispatcher_str += f"if (dtype == {dtype.torch_name})"
            dispatcher_str += " { \\\n"
            dispatcher_str += "    "
            dispatcher_str += f"  {self.name}_{dtype.short_name}(dim, q_tile_size, kv_tile_size, __VA_ARGS__); \\\n"
            dispatcher_str += "    } \\\n"
        dispatcher_str += "    else { \\\n"
        dispatcher_str += '      std::cerr << "Hopper FMHA kernel launch failed!" \\\n'
        dispatcher_str += (
            '                << "'
            + "Hopper FMHA does not support this data type."
            + '" \\\n'
        )
        dispatcher_str += "                << std::endl; \\\n"
        dispatcher_str += "      exit(EXIT_FAILURE); \\\n"
        dispatcher_str += "    } \\\n"
        dispatcher_str += "}();"
        dispatcher_str += "\n\n"
        return dispatcher_str


class HeadDimDispatcher:
    def __init__(self, dtype: DataType):
        self.dtype = dtype
        self.name = f"DISPATCH_HOPPER_FMHA_BACKWARD_{self.dtype.short_name}"

        self.dims: List[int] = []

    def append(self, dim: int):
        self.dims.append(dim)

    def get_dispatcher(self):
        dispatcher_str = ""
        dispatcher_str += (
            f"#define {self.name}(dim, q_tile_size, kv_tile_size, ...) \\\n"
        )
        dispatcher_str += "  [&] { \\\n"
        for i, dim in enumerate(self.dims):
            dispatcher_str += "    "
            if i > 0:
                dispatcher_str += "else "
            dispatcher_str += f"if (dim == {dim})"
            dispatcher_str += " { \\\n"
            dispatcher_str += "    "
            dispatcher_str += f"  {self.name}_headdim{dim}(q_tile_size, kv_tile_size, __VA_ARGS__); \\\n"
            dispatcher_str += "    } \\\n"
        dispatcher_str += "    else { \\\n"
        dispatcher_str += '      std::cerr << "Hopper FMHA kernel launch failed!" \\\n'
        dispatcher_str += (
            '                << "'
            + "Hopper FMHA does not support this data type."
            + '" \\\n'
        )
        dispatcher_str += "                << std::endl; \\\n"
        dispatcher_str += "      exit(EXIT_FAILURE); \\\n"
        dispatcher_str += "    } \\\n"
        dispatcher_str += "}();"
        dispatcher_str += "\n\n"
        return dispatcher_str


class ConfigDispatcher:
    def __init__(
        self,
        dtype: DataType,
        head_dim: int,
    ):
        self.dtype = dtype
        self.head_dim = head_dim

        self.name = (
            f"DISPATCH_HOPPER_FMHA_BACKWARD_{self.dtype.short_name}_headdim{head_dim}"
        )

        self.configs: List = []

    def append(self, config):
        assert len(config) == 2
        assert isinstance(config[0], int)
        assert isinstance(config[1], int)
        self.configs.append(config)

    def get_kernel_instance(self, q_tile_size, kv_tile_size):
        gemm_M = q_tile_size
        gemm_N = kv_tile_size
        gemm_K = self.head_dim
        gemm_shape = (gemm_M, gemm_N, gemm_K)

        supported_configs = SUPPORTED_CONFIGS_BACKWARD[self.dtype.bits][self.head_dim]
        assert (
            gemm_shape in supported_configs
        ), f"{gemm_shape=} not in supported configs {supported_configs=}"

        kernel = HopperFmhaInstance(
            dtype=self.dtype,
            gemm_shape=gemm_shape,
        )

        return kernel

    def get_target_name(self, q_tile_size, kv_tile_size):
        kernel = self.get_kernel_instance(q_tile_size, kv_tile_size)
        return kernel.get_name()

    def get_dispatcher(self):
        dispatcher_str = ""
        dispatcher_str += f"#define {self.name}(q_tile_size, kv_tile_size, ...) \\\n"
        dispatcher_str += "  [&] { \\\n"
        i = 0
        for q_tile_size, kv_tile_size in self.configs:
            dispatcher_str += "    "
            if i > 0:
                dispatcher_str += "else "
            i += 1
            dispatcher_str += "if ("
            dispatcher_str += f"q_tile_size == {q_tile_size}"
            dispatcher_str += " && "
            dispatcher_str += "\\\n"
            dispatcher_str += f"kv_tile_size == {kv_tile_size})"
            dispatcher_str += " { \\\n"

            dispatcher_str += f"  natten::cuda::fmha_hopper::{self.get_target_name(q_tile_size, kv_tile_size)}(__VA_ARGS__); \\\n"

            dispatcher_str += "} \\\n"
        dispatcher_str += "    else { \\\n"
        dispatcher_str += "    "
        dispatcher_str += '      std::cerr << "Hopper FMHA kernel launch failed!" \\\n'
        dispatcher_str += (
            '                << "'
            + "Hopper FMHA got invalid Q tile and KV tile combination."
            + '" \\\n'
        )
        dispatcher_str += "                << std::endl; \\\n"
        dispatcher_str += "      exit(EXIT_FAILURE); \\\n"
        dispatcher_str += "    } \\\n"
        dispatcher_str += "}();"
        dispatcher_str += "\n\n"
        return dispatcher_str


def write_header_file(content, path, namespaces, extra_includes=None):
    extra_includes = extra_includes or []
    header_head = [
        "#pragma once\n",
        "\n\n",
    ]
    header_head += ["#include <iostream> \n"]
    header_head += ["#include <type_traits> \n"]
    header_head += ["#ifdef NATTEN_WITH_CUTLASS\n"]
    header_head += ["#ifdef NATTEN_WITH_HOPPER_FNA\n"]
    for incl in extra_includes:
        header_head += [f"#include <{incl}> \n"]

    for namespace in namespaces:
        header_head += [f"namespace {namespace}", " { \n"]

    header_foot = [
        "\n\n",
    ]
    for namespace in namespaces:
        header_foot += ["} ", f"// namespace {namespace}", " \n"]
    header_foot += [
        "#endif \n",
        "#endif \n",
        "\n",
    ]
    with open(path, "w") as f:
        f.write("".join(header_head))
        f.write(content)
        f.write("".join(header_foot))


def generate_hopper_fmha_kernels(path, num_splits=2):

    SUPPORTED_DTYPES = [
        Half,
        BFloat,
    ]

    HEAD_DIMS = [32, 64, 128]

    CONFIGS = {
        16: {
            32: [
                (64, 128),
                (128, 128),
            ],
            64: [
                (64, 128),
                (128, 128),
            ],
            128: [
                (64, 128),
            ],
        },
    }

    head_dim_dispatchers = []
    config_dispatchers = []
    kernels = []

    dtype_dispatcher = DTypeDispatcher()
    for dtype in SUPPORTED_DTYPES:
        dtype_dispatcher.append(dtype)

        head_dim_dispatcher = HeadDimDispatcher(dtype=dtype)

        for head_dim in HEAD_DIMS:
            head_dim_dispatcher.append(head_dim)

            config_dispatcher = ConfigDispatcher(
                dtype=dtype,
                head_dim=head_dim,
            )

            for q_tile_size, kv_tile_size in CONFIGS[dtype.bits][head_dim]:
                config_dispatcher.append((q_tile_size, kv_tile_size))
                kernels.append(
                    config_dispatcher.get_kernel_instance(q_tile_size, kv_tile_size)
                )
            config_dispatchers.append(config_dispatcher)
        head_dim_dispatchers.append(head_dim_dispatcher)

    #

    path_to_sources = f"{path}/autogen/src/cuda/hopper_fmha_bwd/"
    rel_header = "natten_autogen/cuda/hopper_fmha_bwd/"
    path_to_header_dir = f"{path}/autogen/include/{rel_header}"

    os.makedirs(path_to_sources, exist_ok=False)
    os.makedirs(path_to_header_dir, exist_ok=False)

    path_headers = f"{path_to_header_dir}kernels.h"
    path_dtype = f"{path_to_header_dir}interface.h"
    path_head_dim = f"{path_to_header_dir}dispatch_head_dim.h"
    path_tile_size = f"{path_to_header_dir}dispatch_tile_size.h"

    rel_path_headers = f"{rel_header}kernels.h"
    rel_path_head_dim = f"{rel_header}dispatch_head_dim.h"
    rel_path_tile_size = f"{rel_header}dispatch_tile_size.h"

    dtype_disp = dtype_dispatcher.get_dispatcher()

    head_dim_disp = ""
    for dispatcher in head_dim_dispatchers:
        head_dim_disp += dispatcher.get_dispatcher()

    config_disp = ""
    for dispatcher in config_dispatchers:
        config_disp += dispatcher.get_dispatcher()

    headers = ""
    for kernel in kernels:
        headers += kernel.get_decl()

    assert (
        len(kernels) >= num_splits
    ), f"Generated {len(kernels)} kernels, but got {num_splits=}."
    split_size = len(kernels) // num_splits
    num_splits_with_res = len(kernels) % num_splits
    kernels_emitted = []
    kernels_split = []
    for split_idx in range(num_splits):
        kernel_start_idx = split_size * split_idx + min(num_splits_with_res, split_idx)
        num_kernels_in_split = split_size + (
            1 if split_idx < num_splits_with_res else 0
        )
        kernel_end_idx = kernel_start_idx + num_kernels_in_split
        assert kernel_end_idx <= len(kernels)
        pth_set = set()
        source_list = []
        for kernel_idx in range(kernel_start_idx, kernel_end_idx):
            kernel = kernels[kernel_idx]
            # pth_set.add(kernel.path_to_header)
            source_list.append(kernel)
            kernels_emitted.append(kernel_idx)
        pth_set.add(rel_path_headers)
        write_combined_source_file(
            path_to_sources, f"source_{split_idx}.cu", sorted(pth_set), source_list
        )
        kernels_split.append(source_list)
        # print(f"{split_idx=}, {kernel_start_idx=}, {kernel_end_idx=}, {len(kernels_emitted)=}")
    assert split_idx == num_splits - 1, f"Expected {split_idx=} == {num_splits=} - 1"
    assert len(kernels_emitted) == len(kernels)
    assert sorted(kernels_emitted) == [
        x for x in range(len(kernels))
    ], f"{sorted(kernels_emitted)=}"
    assert all(len(x) > 0 for x in kernels_split)

    namespaces = ["natten", "cuda", "fmha_hopper"]
    cuda_headers = [
        "natten/natten.h",
        "ATen/ATen.h",
        "ATen/cuda/CUDAContext.h",
        "c10/cuda/CUDAGuard.h",
        "c10/cuda/CUDAStream.h",
        "torch/extension.h",
        "natten/natten.h",
        "natten/helpers.h",
        "natten/cuda/fmha_hopper/fmha_backward.cuh",
    ]
    write_header_file(
        dtype_disp, path_dtype, namespaces, cuda_headers + [rel_path_head_dim]
    )
    write_header_file(
        head_dim_disp, path_head_dim, namespaces, cuda_headers + [rel_path_tile_size]
    )
    write_header_file(
        config_disp, path_tile_size, namespaces, cuda_headers + [rel_path_headers]
    )
    write_header_file(headers, path_headers, namespaces, cuda_headers)


def generate_hopper_fmha(output_directory: str, num_splits: int):
    generate_hopper_fmha_kernels(output_directory, num_splits=num_splits)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-o",
        "--output-directory",
        default=DEFAULT_OUTPUT_DIR,
        help="Path to the directory where the auto-generated "
        "kernel instantiations are dumped. "
        f"Default: {DEFAULT_OUTPUT_DIR}",
    )
    parser.add_argument(
        "--num-splits",
        type=int,
        default=1,
        help="Number of source files into which the kernels are split. Default: 1.",
    )
    args = parser.parse_args()
    generate_hopper_fmha(args.output_directory, args.num_splits)



#################################################################################################
# Copyright (c) 2022 - 2026 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################

from typing import Optional

from torch import Tensor

from natten._environment import USE_TORCH_IMPL_DEFAULT
from natten.token_permute.cutlass_impl import (
    can_run_cutlass_tokperm,
    token_permute_cutlass,
    token_unpermute_cutlass,
)
from natten.token_permute.torch_impl import token_permute_torch, token_unpermute_torch
from natten.types import DimensionType
from natten.utils import log
from natten.utils.tuples import ceil_div_tuple, mul_tuple

logger = log.get_logger(__name__)


def token_permute_operation(
    tensor: Tensor,
    tile_shape: DimensionType,
    dilation: Optional[DimensionType] = None,
    flip_tiled_dims: bool = True,
    use_torch: bool = USE_TORCH_IMPL_DEFAULT,
) -> tuple[Tensor, DimensionType, DimensionType]:
    if tensor.dim() not in [4, 5, 6]:
        raise ValueError(
            "Expected 4D, 5D, or 6D tensor (corresponding to NA1D, 2D, 3D), "
            f"got {tensor.dim()}D input."
        )

    na_dim = tensor.dim() - 3
    assert na_dim in [1, 2, 3]

    if len(tile_shape) != na_dim:
        raise ValueError(
            f"Expected {na_dim}D tiler for NA{na_dim}D, " f"got {tile_shape=}."
        )

    if dilation is not None and len(dilation) != na_dim:
        raise ValueError(
            f"Expected {na_dim}D dilation for NA{na_dim}D, " f"got {dilation=}."
        )

    dilation_: DimensionType = dilation or tuple(1 for _ in range(na_dim))  # type: ignore[assignment]

    tensor = tensor.contiguous()
    batch, *token_layout_, heads, dim = tensor.shape
    token_layout: DimensionType = tuple(x for x in token_layout_)  # type: ignore[assignment]

    token_layout_post_dilation: DimensionType = mul_tuple(ceil_div_tuple(ceil_div_tuple(token_layout, tile_shape), dilation_), tile_shape)  # type: ignore[assignment]

    if not use_torch and can_run_cutlass_tokperm(tensor):
        output = token_permute_cutlass(
            tensor,
            tile_shape=tile_shape,
            dilation=dilation_,
            flip_tiled_dims=flip_tiled_dims,
        )
    else:
        output = token_permute_torch(
            tensor,
            tile_shape=tile_shape,
            dilation=dilation_,
            flip_tiled_dims=flip_tiled_dims,
        )

    return output, token_layout, token_layout_post_dilation


def token_unpermute_operation(
    tensor: Tensor,
    token_layout_shape: DimensionType,
    tile_shape: DimensionType,
    dilation: Optional[DimensionType] = None,
    flip_tiled_dims: bool = True,
    use_torch: bool = USE_TORCH_IMPL_DEFAULT,
) -> Tensor:
    if tensor.dim() != 4:
        raise ValueError(f"Expected flattened 4D tensor, got {tensor.dim()}D input.")

    na_dim = len(token_layout_shape)

    if len(tile_shape) != na_dim:
        raise ValueError(
            f"Expected {na_dim}D tiler for NA{na_dim}D, " f"got {tile_shape=}."
        )

    if dilation is not None and len(dilation) != na_dim:
        raise ValueError(
            f"Expected {na_dim}D dilation for NA{na_dim}D, " f"got {dilation=}."
        )

    dilation_: DimensionType = dilation or tuple(1 for _ in range(na_dim))  # type: ignore[assignment]

    tensor = tensor.contiguous()
    if not use_torch and can_run_cutlass_tokperm(tensor):
        output = token_unpermute_cutlass(
            tensor,
            token_layout_shape,
            tile_shape=tile_shape,
            dilation=dilation_,
            flip_tiled_dims=flip_tiled_dims,
        )
    else:
        output = token_unpermute_torch(
            tensor,
            token_layout_shape,
            tile_shape=tile_shape,
            dilation=dilation_,
            flip_tiled_dims=flip_tiled_dims,
        )

    return output



# Copyright (c) 2022 - 2026 Ali Hassani.
#
# This script is intended to emit fused kernel instantiations into
# a variable number of source files, generate appropriate headers
# and a single dispatcher interface, which will be used by the
# NATTEN API to call the kernels.
#
# NOTE: these scripts are heavily under-documented, and
# overly-repetitive, and will be replaced in future PRs.
# Please use it with caution.

import argparse
import os
from typing import List


DEFAULT_OUTPUT_DIR = "csrc/"


KERNEL_DECL_TEMPLATE = """
void {kernel_name}(
      void* ptr_Q,
      void* ptr_K,
      void* ptr_V,
      void* ptr_O,
      void* ptr_LSE,
      int batch_size,
      int seqlen,
      int heads,
      int heads_kv,
      int dim,
      int dim_value,
      int num_additional_kv,
      {DimType} qkv_shape,
      {DimType} window_size,
      {DimType} stride,
      {DimType} dilation,
      float attn_scale,
      cudaStream_t stream);
"""


KERNEL_IMPL_TEMPLATE = """
void {kernel_name}(
      void* ptr_Q,
      void* ptr_K,
      void* ptr_V,
      void* ptr_O,
      void* ptr_LSE,
      int batch_size,
      int seqlen,
      int heads,
      int heads_kv,
      int dim,
      int dim_value,
      int num_additional_kv,
      {DimType} qkv_shape,
      {DimType} window_size,
      {DimType} stride,
      {DimType} dilation,
      float attn_scale,
      cudaStream_t stream) {{

  using Causal = {Causal};

  fna_reference_forward(
    static_cast<{dtype}*>(ptr_Q),
    static_cast<{dtype}*>(ptr_K),
    static_cast<{dtype}*>(ptr_V),
    static_cast<{dtype}*>(ptr_O),
    static_cast<float*>(ptr_LSE),
    batch_size,
    seqlen,
    heads,
    heads_kv,
    dim,
    dim_value,
    num_additional_kv,
    qkv_shape,
    window_size,
    stride,
    dilation,
    Causal{{}},
    attn_scale,
    stream);
}}
"""


KERNEL_BWD_DECL_TEMPLATE = """
void {kernel_name}(
      void* ptr_Q,
      void* ptr_K,
      void* ptr_V,
      void* ptr_O,
      void* ptr_DO,
      void* ptr_DQ,
      void* ptr_DK,
      void* ptr_DV,
      void* ptr_LSE,
      int batch_size,
      int seqlen,
      int heads,
      int heads_kv,
      int dim,
      int dim_value,
      int num_additional_kv,
      {DimType} qkv_shape,
      {DimType} window_size,
      {DimType} stride,
      {DimType} dilation,
      float attn_scale,
      cudaStream_t stream);
"""


KERNEL_BWD_IMPL_TEMPLATE = """
void {kernel_name}(
      void* ptr_Q,
      void* ptr_K,
      void* ptr_V,
      void* ptr_O,
      void* ptr_DO,
      void* ptr_DQ,
      void* ptr_DK,
      void* ptr_DV,
      void* ptr_LSE,
      int batch_size,
      int seqlen,
      int heads,
      int heads_kv,
      int dim,
      int dim_value,
      int num_additional_kv,
      {DimType} qkv_shape,
      {DimType} window_size,
      {DimType} stride,
      {DimType} dilation,
      float attn_scale,
      cudaStream_t stream) {{

  using Causal = {Causal};

  fna_reference_backward(
    static_cast<{dtype}*>(ptr_Q),
    static_cast<{dtype}*>(ptr_K),
    static_cast<{dtype}*>(ptr_V),
    static_cast<{dtype}*>(ptr_O),
    static_cast<{dtype}*>(ptr_DO),
    static_cast<{dtype}*>(ptr_DQ),
    static_cast<{dtype}*>(ptr_DK),
    static_cast<{dtype}*>(ptr_DV),
    static_cast<float*>(ptr_LSE),
    batch_size,
    seqlen,
    heads,
    heads_kv,
    dim,
    dim_value,
    num_additional_kv,
    qkv_shape,
    window_size,
    stride,
    dilation,
    Causal{{}},
    attn_scale,
    stream);
}}
"""


class DataType:
    def __init__(self, name, short_name, torch_name, bits):
        self.name = name
        self.bits = bits
        self.short_name = short_name
        self.torch_name = torch_name


Float = DataType("float", "float32", "torch::kFloat32", 32)
Half = DataType("cutlass::half_t", "float16", "torch::kFloat16", 16)
BFloat = DataType("cutlass::bfloat16_t", "bfloat16", "torch::kBFloat16", 16)


def iterable_to_static_cute_tuple(shape_in) -> str:
    shape = ", ".join([f"cute::Int<{x}>" for x in shape_in])
    return f"cute::tuple<{shape}>"


def get_dim_type(na_dim: int) -> str:
    shape = ", ".join(["int" for _ in range(na_dim)])
    return f"cute::tuple<{shape}>"


class ReferenceFnaInstance:
    def __init__(
        self,
        na_dim: int,
        dtype: DataType,
        causal: tuple,
        is_backward: bool,
    ):
        assert 0 < na_dim <= 3
        assert na_dim == len(causal)
        self.na_dim = na_dim
        self.causal = causal
        self.dtype = dtype
        self.is_backward = is_backward

    def get_causal_cute(self) -> str:
        consts = ", ".join(
            ["cute::true_type" if c else "cute::false_type" for c in self.causal]
        )
        return f"cute::tuple<{consts}>"

    def get_name(self) -> str:
        backward_str = "" if not self.is_backward else "_backward"
        name = f"reference_fna{self.na_dim}d{backward_str}"
        name += f"_{self.dtype.short_name}"
        name += "_causal" + "x".join(["1" if c else "0" for c in self.causal])
        return name

    def get_decl(self) -> str:
        return (
            KERNEL_BWD_DECL_TEMPLATE if self.is_backward else KERNEL_DECL_TEMPLATE
        ).format(
            kernel_name=self.get_name(),
            DimType=get_dim_type(self.na_dim),
            dtype=self.dtype.name,
        )

    def get_impl(self) -> str:
        return (
            KERNEL_BWD_IMPL_TEMPLATE if self.is_backward else KERNEL_IMPL_TEMPLATE
        ).format(
            kernel_name=self.get_name(),
            DimType=get_dim_type(self.na_dim),
            Causal=self.get_causal_cute(),
            dtype=self.dtype.name,
        )


def write_combined_source_file(path, filename, headers, kernels):
    source_head = []
    source_head += ["#ifdef NATTEN_WITH_CUTLASS\n"]

    source_head += ["#include <cuda_runtime.h>\n"]
    source_head += ["#include <iostream>\n"]

    source_head += ["#include <ATen/ATen.h>\n"]
    source_head += ["#include <ATen/cuda/CUDAContext.h>\n"]
    source_head += ["#include <c10/cuda/CUDAGuard.h>\n"]
    source_head += ["#include <c10/cuda/CUDAStream.h>\n"]
    source_head += ["#include <torch/extension.h>\n"]

    source_head += ["#include <natten/natten.h>\n"]
    source_head += ["#include <natten/helpers.h>\n"]

    source_head += ["#include <natten/cuda/reference/fna_reference_forward.hpp>\n"]
    source_head += ["#include <natten/cuda/reference/fna_reference_backward.hpp>\n"]

    for header in headers:
        source_head += [f"#include <{header}>\n"]

    source_head += ["namespace natten { \n"]
    source_head += ["namespace cuda { \n"]
    source_head += ["namespace reference { \n\n"]

    source_head = "".join(source_head)

    source_body = []
    for kernel in kernels:
        source_body += "\n\n" + kernel.get_impl() + "\n\n"
    source_body = "".join(source_body)

    source_foot = "".join(
        [
            "} // namespace reference \n",
            "} // namespace cuda \n",
            "} // namespace natten \n",
            "#endif \n",
            "\n",
        ]
    )
    filename = f"{path}/{filename}"
    with open(filename, "w") as f:
        f.write(source_head)
        f.write(source_body)
        f.write(source_foot)


class NaDimDispatcher:
    def __init__(self, is_backward: bool):
        fwd_bwd_str = "BACKWARD" if is_backward else "FORWARD"
        self.name = f"DISPATCH_REFERENCE_FNA_{fwd_bwd_str}"
        self.dims: List[int] = []
        self.is_backward = is_backward

    def append(self, na_dim: int):
        self.dims.append(na_dim)

    def get_dispatcher(self):
        dispatcher_str = ""
        dispatcher_str += f"#define {self.name}(rank, dtype, is_causal, ...) \\\n"
        dispatcher_str += "  [&] { \\\n"
        for i, na_dim in enumerate(self.dims):
            dispatcher_str += "    "
            if i > 0:
                dispatcher_str += "else "
            dispatcher_str += f"if constexpr (rank == {na_dim})"
            dispatcher_str += " { \\\n"
            dispatcher_str += "    "
            dispatcher_str += (
                f"  {self.name}_{na_dim}D(dtype, is_causal, __VA_ARGS__); \\\n"
            )
            dispatcher_str += "    } \\\n"
        dispatcher_str += "    else { \\\n"
        dispatcher_str += (
            '      std::cerr << "Reference FNA kernel launch failed!" \\\n'
        )
        dispatcher_str += (
            '                << "' + "NATTEN only supports NA1D, 2D, and 3D!" + '" \\\n'
        )
        dispatcher_str += "                << std::endl; \\\n"
        dispatcher_str += "      exit(EXIT_FAILURE); \\\n"
        dispatcher_str += "    } \\\n"
        dispatcher_str += "}();"
        dispatcher_str += "\n\n"
        return dispatcher_str


class DTypeDispatcher:
    def __init__(self, is_backward: bool, na_dim: int):
        self.dtypes: List[DataType] = []
        self.na_dim = na_dim
        fwd_bwd_str = "BACKWARD" if is_backward else "FORWARD"
        self.name = f"DISPATCH_REFERENCE_FNA_{fwd_bwd_str}_{self.na_dim}D"
        self.is_backward = is_backward

    def append(self, dtype: DataType):
        self.dtypes.append(dtype)

    def get_dispatcher(self):
        dispatcher_str = ""
        dispatcher_str += f"#define {self.name}(dtype, is_causal, ...) \\\n"
        dispatcher_str += "  [&] { \\\n"
        for i, dtype in enumerate(self.dtypes):
            dispatcher_str += "    "
            if i > 0:
                dispatcher_str += "else "
            dispatcher_str += f"if (dtype == {dtype.torch_name})"
            dispatcher_str += " { \\\n"
            dispatcher_str += "    "
            dispatcher_str += (
                f"  {self.name}_{dtype.short_name}(is_causal, __VA_ARGS__); \\\n"
            )
            dispatcher_str += "    } \\\n"
        dispatcher_str += "    else { \\\n"
        dispatcher_str += (
            '      std::cerr << "Reference FNA kernel launch failed!" \\\n'
        )
        dispatcher_str += (
            '                << "'
            + f"Reference FNA-{self.na_dim}D does not support this data type."
            + '" \\\n'
        )
        dispatcher_str += "                << std::endl; \\\n"
        dispatcher_str += "      exit(EXIT_FAILURE); \\\n"
        dispatcher_str += "    } \\\n"
        dispatcher_str += "}();"
        dispatcher_str += "\n\n"
        return dispatcher_str


class CausalMaskDispatcher:
    def __init__(
        self,
        is_backward: bool,
        na_dim: int,
        dtype: DataType,
    ):
        self.na_dim = na_dim
        self.dtype = dtype

        fwd_bwd_str = "BACKWARD" if is_backward else "FORWARD"
        self.name = f"DISPATCH_REFERENCE_FNA_{fwd_bwd_str}_{self.na_dim}D_{self.dtype.short_name}"
        self.is_backward = is_backward

        self.cms: List = []

    def append(self, cm):
        assert len(cm) == self.na_dim, f"{cm} incompatible for {self.na_dim}D NA."
        self.cms.append(cm)

    def get_kernel_instance(self, cm):
        return ReferenceFnaInstance(
            na_dim=self.na_dim,
            dtype=self.dtype,
            causal=cm,
            is_backward=self.is_backward,
        )

    def get_target_name(self, cm):
        kernel = self.get_kernel_instance(cm)
        return kernel.get_name()

    def get_dispatcher(self):
        dispatcher_str = ""
        dispatcher_str += f"#define {self.name}(is_causal, ...) \\\n"
        dispatcher_str += "  [&] { \\\n"
        i = 0
        for cm in self.cms:
            dispatcher_str += "    "
            if i > 0:
                dispatcher_str += "else "
            i += 1
            dispatcher_str += "if ("
            for dim in range(self.na_dim):
                dispatcher_str += (
                    f"cute::get<{dim}>(is_causal)"
                    if cm[dim]
                    else f"not cute::get<{dim}>(is_causal)"
                )
                if dim != self.na_dim - 1:
                    dispatcher_str += " && "
            dispatcher_str += ")"
            dispatcher_str += " { \\\n"
            dispatcher_str += "    "

            dispatcher_str += f"    natten::cuda::reference::{self.get_target_name(cm)}(__VA_ARGS__); \\\n"

            dispatcher_str += "    } \\\n"
        dispatcher_str += "    else { \\\n"
        dispatcher_str += "    "
        dispatcher_str += (
            '      std::cerr << "Reference FNA kernel launch failed!" \\\n'
        )
        dispatcher_str += (
            '                << "'
            + "Causal mask dispatcher got invalid causal mask!"
            + '" \\\n'
        )
        dispatcher_str += "                << std::endl; \\\n"
        dispatcher_str += "      exit(EXIT_FAILURE); \\\n"
        dispatcher_str += "    } \\\n"
        dispatcher_str += "}();"
        dispatcher_str += "\n\n"
        return dispatcher_str


def write_header_file(content, path, namespaces, extra_includes=None):
    extra_includes = extra_includes or []
    header_head = [
        "#pragma once\n",
        "\n\n",
    ]
    header_head += ["#include <iostream> \n"]
    header_head += ["#include <type_traits> \n"]
    header_head += ["#ifdef NATTEN_WITH_CUTLASS\n"]
    for incl in extra_includes:
        header_head += [f"#include <{incl}> \n"]

    for namespace in namespaces:
        header_head += [f"namespace {namespace}", " { \n"]

    header_foot = [
        "\n\n",
    ]
    for namespace in namespaces:
        header_foot += ["} ", f"// namespace {namespace}", " \n"]
    header_foot += [
        "#endif \n",
        "\n",
    ]
    with open(path, "w") as f:
        f.write("".join(header_head))
        f.write(content)
        f.write("".join(header_foot))


def generate_reference_fna_kernels(path, num_splits=2):

    NA_DIMS = [1, 2, 3]

    SUPPORTED_DTYPES = [
        Float,
        Half,
        BFloat,
    ]

    CAUSAL_MASKS = {
        1: [(False,), (True,)],
        2: [(False, False), (False, True), (True, False), (True, True)],
        3: [
            (False, False, False),
            (False, False, True),
            (False, True, False),
            (False, True, True),
            (True, False, False),
            (True, False, True),
            (True, True, False),
            (True, True, True),
        ],
    }

    dtype_dispatchers = []
    cm_dispatchers = []
    kernels = []

    rank_dispatcher = NaDimDispatcher(is_backward=False)
    rank_dispatcher_bwd = NaDimDispatcher(is_backward=True)
    for na_dim in NA_DIMS:
        rank_dispatcher.append(na_dim)
        rank_dispatcher_bwd.append(na_dim)

        for is_backward in [False, True]:
            dtype_dispatcher = DTypeDispatcher(is_backward=is_backward, na_dim=na_dim)
            for dtype in SUPPORTED_DTYPES:
                dtype_dispatcher.append(dtype)

                cm_dispatcher = CausalMaskDispatcher(
                    is_backward=is_backward, na_dim=na_dim, dtype=dtype
                )
                for cm in CAUSAL_MASKS[na_dim]:
                    cm_dispatcher.append(cm)
                    kernels.append(cm_dispatcher.get_kernel_instance(cm))

                cm_dispatchers.append(cm_dispatcher)
            dtype_dispatchers.append(dtype_dispatcher)

    #

    path_to_sources = f"{path}/autogen/src/cuda/reference/"
    rel_header = "natten_autogen/cuda/reference/"
    path_to_header_dir = f"{path}/autogen/include/{rel_header}"

    os.makedirs(path_to_sources, exist_ok=False)
    os.makedirs(path_to_header_dir, exist_ok=False)

    path_headers = f"{path_to_header_dir}kernels.h"
    path_rank = f"{path_to_header_dir}interface.h"
    path_dtype = f"{path_to_header_dir}dispatch_dtype.h"
    path_cm = f"{path_to_header_dir}dispatch_cm.h"

    rel_path_headers = f"{rel_header}kernels.h"
    rel_path_dtype = f"{rel_header}dispatch_dtype.h"
    rel_path_cm = f"{rel_header}dispatch_cm.h"

    rank_disp = rank_dispatcher.get_dispatcher() + rank_dispatcher_bwd.get_dispatcher()

    dtype_disp = ""
    for dispatcher in dtype_dispatchers:
        dtype_disp += dispatcher.get_dispatcher()

    cm_disp = ""
    for dispatcher in cm_dispatchers:
        cm_disp += dispatcher.get_dispatcher()

    headers = ""
    for kernel in kernels:
        headers += kernel.get_decl()

    assert (
        len(kernels) >= num_splits
    ), f"Generated {len(kernels)} kernels, but got {num_splits=}."
    split_size = len(kernels) // num_splits
    num_splits_with_res = len(kernels) % num_splits
    kernels_emitted = []
    kernels_split = []
    for split_idx in range(num_splits):
        kernel_start_idx = split_size * split_idx + min(num_splits_with_res, split_idx)
        num_kernels_in_split = split_size + (
            1 if split_idx < num_splits_with_res else 0
        )
        kernel_end_idx = kernel_start_idx + num_kernels_in_split
        assert kernel_end_idx <= len(kernels)
        pth_set = set()
        source_list = []
        for kernel_idx in range(kernel_start_idx, kernel_end_idx):
            kernel = kernels[kernel_idx]
            source_list.append(kernel)
            kernels_emitted.append(kernel_idx)
        pth_set.add(rel_path_headers)
        write_combined_source_file(
            path_to_sources, f"source_{split_idx}.cu", sorted(pth_set), source_list
        )
        kernels_split.append(source_list)
        # print(f"{split_idx=}, {kernel_start_idx=}, {kernel_end_idx=}, {len(kernels_emitted)=}")
    assert split_idx == num_splits - 1, f"Expected {split_idx=} == {num_splits=} - 1"
    assert len(kernels_emitted) == len(kernels)
    assert sorted(kernels_emitted) == [
        x for x in range(len(kernels))
    ], f"{sorted(kernels_emitted)=}"
    assert all(len(x) > 0 for x in kernels_split)

    namespaces = ["natten", "cuda", "reference"]
    cuda_headers = [
        "natten/natten.h",
        "ATen/ATen.h",
        "ATen/cuda/CUDAContext.h",
        "c10/cuda/CUDAGuard.h",
        "c10/cuda/CUDAStream.h",
        "torch/extension.h",
        "natten/natten.h",
        "natten/helpers.h",
        "natten/cuda/reference/fna_reference_forward.hpp",
        "natten/cuda/reference/fna_reference_backward.hpp",
    ]
    write_header_file(rank_disp, path_rank, namespaces, cuda_headers + [rel_path_dtype])
    write_header_file(dtype_disp, path_dtype, namespaces, cuda_headers + [rel_path_cm])
    write_header_file(cm_disp, path_cm, namespaces, cuda_headers + [rel_path_headers])
    write_header_file(headers, path_headers, namespaces, cuda_headers)


def generate_reference_fna(output_directory: str, num_splits: int):
    generate_reference_fna_kernels(output_directory, num_splits=num_splits)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-o",
        "--output-directory",
        default=DEFAULT_OUTPUT_DIR,
        help="Path to the directory where the auto-generated "
        "kernel instantiations are dumped. "
        f"Default: {DEFAULT_OUTPUT_DIR}",
    )
    parser.add_argument(
        "--num-splits",
        type=int,
        default=2,
        help="Number of source files into which the kernels are split. Default: 2.",
    )
    args = parser.parse_args()
    generate_reference_fna(args.output_directory, args.num_splits)
